# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import abc
import ast
from typing import List

import logging

from kqcircuits.defaults import default_faces
from kqcircuits.elements.airbridges.airbridge import Airbridge
from kqcircuits.elements.element import Element
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node
from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.port import Port, InternalPort, EdgePort
from kqcircuits.util.geometry_helper import region_with_merged_polygons, region_with_merged_points, \
    match_points_on_edges
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.simulations.export.util import find_edge_from_point_in_cell
from kqcircuits.simulations.export.util import get_enclosing_polygon
from kqcircuits.util.groundgrid import make_grid
from kqcircuits.junctions.sim import Sim


simulation_layer_dict = dict()


def get_simulation_layer_by_name(layer_name):
    """Returns layer info of given name. If layer doesn't exist, a new layer is created.
    New layers are created with data type = 0 and layer numbering starts from 1000."""
    if layer_name not in simulation_layer_dict:
        simulation_layer_dict[layer_name] = pya.LayerInfo(len(simulation_layer_dict) + 1000, 0, layer_name)
    return simulation_layer_dict[layer_name]


@add_parameters_from(Element)
@add_parameters_from(Sim, "junction_total_length")
class Simulation:
    """Base class for simulation geometries.

    Generally, this class is intended to be subclassed by a specific simulation implementation; the
    implementation defines the simulation geometry ad ports in `build`.

    A convenience class method `Simulation.from_cell` is provided to create a Simulation from an
    existing cell. In this case no ports will be added.

    Basically, 3D layout is built of substrates, which are separated from each other by vacuum boxes, however, this rule
    is modifiable by setting substrate and vacuum thicknesses to zero. In principle, one can stack faces on any of the
    imaginable surface of a substrate. If substrate or vacuum thickness is set to zero, then there can be two
    faces touching each other. Faces can be stacked on bottom or top surface of the substrates.

    Number of substrate and vacuum boxes are determined with parameters face_stack and lower_box_height:
    - If lower_box_height > 0, there will be a vacuum box below the lowest substrate, and the counting of faces will
    start from the bottom surface of the lowest substrate. Otherwise, the lowest substrate will be below the lowest
    vacuum box, and the counting of faces will start from the top surface of the lowest substrate.
    - Length of face_stack list describes how many surfaces are taken into account in the simulation.

    The terms in the face_stack indicate in which order the klayout faces are stacked in the 3D layout. Faces are
    counted from the lowest substrate surface to the highest. One can also introduce face_stack as list of lists.
    If a term in face_stack is a list, then all the faces given in the list are piled up on the corresponding surface
    in the respective order. That means, the first term in the inner list indicates the face that is closest to the
    surface of the substrate. One can use empty list in face_stack to leave certain surface
    without metallization.

    Heights of substrates (substrate_height) and vacuum boxes between surfaces (chip_distance) can be determined
    individually from bottom to top or with single value. Any of the heights can be left zero, to indicate that there
    is no vacuum between the substrates or substrate between the vacuum boxes. Also, the metal thickness (metal_height)
    can be set to zero, but that means the metal layer is modelled as infinitely thin sheet. The insulator dielectric
    can be set on any metal layer if non-zero dielectric_height is given.
    """
    # The samples below show, how the layout is changed according to the parameters. Number of faces is unlimited:
    #
    # len(face_stack) = 1             len(face_stack) = 2             len(face_stack) = 2
    # lower_box_height = 0            lower_box_height = 0            lower_box_height > 0
    # |-------------------------|     |-------------------------|     |-------------------------|
    # |                         |     |/////////////////////////|     |                         |
    # |                         |     |///substrate_height[1]///|     |    upper_box_height     |
    # |    upper_box_height     |     |/////////////////////////|     |                         |
    # |                         |     |----- face_stack[1] -----|     |----- face_stack[1] -----|
    # |                         |     |                         |     |/////////////////////////|
    # |----- face_stack[0] -----|     |    chip_distance[0]     |     |///substrate_height[0]///|
    # |/////////////////////////|     |                         |     |/////////////////////////|
    # |/////////////////////////|     |----- face_stack[0] -----|     |----- face_stack[0] -----|
    # |///substrate_height[0]///|     |/////////////////////////|     |                         |
    # |/////////////////////////|     |///substrate_height[0]///|     |    lower_box_height     |
    # |/////////////////////////|     |/////////////////////////|     |                         |
    # |-------------------------|     |-------------------------|     |-------------------------|

    LIBRARY_NAME = None  # This is needed by some methods inherited from Element.

    # Metadata associated with simulation
    ports: List[Port]

    # Parameters
    box = Param(pdt.TypeShape, "Boundary box", pya.DBox(pya.DPoint(0, 0), pya.DPoint(10000, 10000)))
    ground_grid_box = Param(pdt.TypeShape, "Border", pya.DBox(pya.DPoint(0, 0), pya.DPoint(10000, 10000)))
    with_grid = Param(pdt.TypeBoolean, "Make ground plane grid", False)
    name = Param(pdt.TypeString, "Name of the simulation", "Simulation")

    use_ports = Param(pdt.TypeBoolean, "Turn off to disable all ports (for debugging)", True)
    use_internal_ports = Param(pdt.TypeBoolean, "Use internal (lumped) ports. The alternative is wave ports.", True)
    port_size = Param(pdt.TypeDouble, "Width and height of wave ports", 400.0, unit="µm",
                      docstring="The port size can also be set as a list specifying the extensions from the center of "
                                "the port to left, right, down and up, respectively.")

    upper_box_height = Param(pdt.TypeDouble, "Height of vacuum above top substrate", 1000.0, unit="µm")
    lower_box_height = Param(pdt.TypeDouble, "Height of vacuum below bottom substrate", 0, unit="µm",
                             docstring="Set > 0 to start face counting from substrate bottom layer.")
    fixed_level_stackup = Param(pdt.TypeBoolean, "Use fixed level multi-face stack-up", True)
    face_stack = Param(pdt.TypeList, "Face IDs on the substrate surfaces from bottom to top", ["1t1"],
                       docstring="The parameter can be set as list of lists to enable multi-face stack-up on substrate "
                                 "surfaces. Set term to empty list to not have metal on the surface.")
    substrate_height = Param(pdt.TypeList, "Height of the substrates", [550.0, 375.0], unit="[µm]",
                             docstring="The value can be scalar or list of scalars. Set as list to use individual "
                                       "substrate heights from bottom to top.")
    substrate_box = Param(pdt.TypeList, "x and y dimensions of substrates", [None],
                          docstring="Set as a list of pya.DBox objects to give substrate dimensions individually from "
                                    "bottom to top. If a value in the list is not pya.DBox object, the corresponding"
                                    "substrate covers fully the general boundary box.")
    substrate_material = Param(pdt.TypeList, "Material of the substrates.", ['silicon'],
                               docstring="Value can be string or list of strings. Use only keywords introduced in "
                                         "material_dict. Set as list to use individual materials from bottom to top.")
    material_dict = Param(pdt.TypeString, "Dictionary of dielectric materials", "{'silicon': {'permittivity': 11.45}}",
                          docstring="Material property keywords follow Ansys Electromagnetics property names. "
                                    "For example 'permittivity', 'dielectric_loss_tangent', etc.")
    chip_distance = Param(pdt.TypeList, "Height of vacuum between two substrates", [8.0], unit="[µm]",
                          docstring="The value can be scalar or list of scalars. Set as list to use individual chip "
                                    "distances from bottom to top.")
    ground_metal_height = Param(pdt.TypeDouble, "Height of the grounded metal (in Xsection tool)", 0.2, unit="µm",
                                docstring="Only used in Xsection tool and doesn't affect the 3D model")
    signal_metal_height = Param(pdt.TypeDouble, "Height of the trace metal (in Xsection tool)", 0.2, unit="µm",
                                docstring="Only used in Xsection tool and doesn't affect the 3D model")

    airbridge_height = Param(pdt.TypeDouble, "Height of airbridges.", 3.4, unit="µm")
    metal_height = Param(pdt.TypeList, "Height of metal sheet on each face.", [0.0], unit="µm")
    dielectric_height = Param(pdt.TypeList, "Height of insulator dielectric on each face.", [0.0], unit="µm")
    dielectric_material = Param(pdt.TypeList, "Material of insulator dielectric on each face.", ['silicon'], unit="µm",
                                docstring="Use only keywords introduced in material_dict.")

    waveguide_length = Param(pdt.TypeDouble, "Length of waveguide stubs or distance between couplers and waveguide "
                                             "turning point", 100, unit="µm")
    over_etching = Param(pdt.TypeDouble, "Expansion of metal gaps (negative to shrink the gaps).", 0, unit="μm")
    vertical_over_etching = Param(pdt.TypeDouble, "Vertical over-etching into substrates at gaps.", 0, unit="μm")
    hollow_tsv = Param(pdt.TypeBoolean, "Make TSVs hollow with vacuum inside and thin metal boundary.", False)

    metal_edge_region_dimensions = Param(pdt.TypeList, "Dimensions of metal edge region", [], unit="µm",
                                         docstring="Metal edge region is disabled if the list is empty. The terms in "
                                                   "the list correspond to expansion dimensions into direction of gap, "
                                                   "vacuum, metal, and substrate, respectively. The implementation "
                                                   "uses the modulo operator in indexing, so one can set the list as "
                                                   "[r] meaning that r is the expansion to all directions, or set as "
                                                   "[r_lat, r_vert] to separate lateral and vertical expansions.")
    tls_layer_thickness = Param(pdt.TypeList, "Thickness of TLS interface layers (MA, MS, and SA, respectively)", [0.0],
                                unit="µm")
    tls_layer_material = Param(pdt.TypeList, "Materials of TLS interface layers (MA, MS, and SA, respectively)", None,
                               docstring="Use None to create non-model layers. Otherwise, use only keywords "
                                         "introduced in material_dict.")
    tls_sheet_approximation = Param(pdt.TypeBoolean, "Approximate TLS interface layers as sheets", False)

    minimum_point_spacing = Param(pdt.TypeDouble, "Tolerance for merging adjacent points in polygon", 0.01, unit="µm")
    polygon_tolerance = Param(pdt.TypeDouble, "Tolerance for merging adjacent polygons in a layer", 0.004, unit="µm")

    extra_json_data = Param(pdt.TypeNone, "Extra data in dict form to store in resulting JSON", None,
        docstring="This field may be used to store 'virtual' parameters useful for your simulations")

    def __init__(self, layout, **kwargs):
        """Initialize a Simulation.

        The initializer parses parameters, creates a top cell, and then calls `self.build` to create
        the simulation geometry, followed by `self.create_simulation_layers` to process the geometry
        so it is ready for exporting.

        Args:
            layout: the layout on which to create the simulation

        Keyword arguments:
            `**kwargs`:
                Any parameter can be passed as a keyword argument.

                In addition, `cell` can be passed as keyword argument. If `cell` is supplied, it will be
                used as the top cell for the simulation. Otherwise, a new cell will be created. See
                `Simulation.from_cell` for creating simulations from existing cells.
        """
        self.refpoints = {}
        if layout is None or not isinstance(layout, pya.Layout):
            error_text = "Cannot create simulation with invalid or nil layout."
            error = ValueError(error_text)
            logging.error(error_text)
            raise error

        self.layout = layout
        schema = type(self).get_schema()

        # Apply kwargs or default value
        # TODO: Validation? Could reuse the validation for Element
        for parameter, item in schema.items():
            if parameter in kwargs:
                setattr(self, parameter, kwargs[parameter])
            else:
                setattr(self, parameter, item.default)

        self.name = self.name.replace(" ", "").replace(",", "__")  # no spaces or commas in filenames

        self.ports = []
        if 'cell' in kwargs:
            self.cell = kwargs['cell']
        else:
            self.cell = layout.create_cell(self.name)

        self.layers = dict()
        self.build()
        self.create_simulation_layers()

    @classmethod
    def from_cell(cls, cell, margin=300, grid_size=1, **kwargs):
        """Create a Simulation from an existing cell.

        Arguments:
            cell: existing top cell for the Simulation
            margin: distance (μm) to expand the simulation box (ground plane) around the bounding
                box of the cell If the `box` keyword argument is given, margin is ignored.
            grid_size: size of the simulation box will be rounded to this resolution
                If the `box` keyword argument is given, grid_size is ignored.
            `**kwargs`: any simulation parameters passed

        Returns:
            Simulation instance
        """
        extra_kwargs = {}

        if 'box' not in kwargs:
            box = cell.dbbox().enlarge(margin, margin)
            if grid_size > 0:
                box.left = round(box.left / grid_size) * grid_size
                box.right = round(box.right / grid_size) * grid_size
                box.bottom = round(box.bottom / grid_size) * grid_size
                box.top = round(box.top / grid_size) * grid_size
            extra_kwargs['box'] = box

        return cls(cell.layout(), cell=cell, **kwargs, **extra_kwargs)

    @abc.abstractmethod
    def build(self):
        """Build simulation geometry.

        This method is to be overridden, and the overriding method should create the geometry to be
        simulated and add any ports to `self.ports`.
        """
        return

    # Inherit specific methods from Element
    get_schema = classmethod(Element.get_schema.__func__)
    face = Element.face
    insert_cell = Element.insert_cell
    get_refpoints = Element.get_refpoints
    add_element = Element.add_element
    get_layer = Element.get_layer
    pcell_params_by_name = Element.pcell_params_by_name

    def face_stack_list_of_lists(self):
        """Return self.face_stack forced to be list of lists"""
        return [f if isinstance(f, list) else [f] for f in self.face_stack]

    def _face_box(self, i):
        """ Return the boundary box for given face number.
        Boundary box is the intersection of the global boundary box and the substrate x-y-dimension box.
        """
        box = self.ith_value(self.substrate_box, (i + int(self.lower_box_height <= 0)) // 2)
        return self.box & box if isinstance(box, pya.DBox) else self.box

    @staticmethod
    def ith_value(list_or_constant, i):
        """ Helper function to return value from list or constant corresponding to the ordinal number i.
        Too short lists are extended by duplicating the last value of the list.
        """
        if isinstance(list_or_constant, list):
            if i < len(list_or_constant):
                return list_or_constant[i]  # return ith term of the list
            return list_or_constant[-1]  # return last term of the list
        return list_or_constant  # return constant value

    def face_z_levels(self):
        """ Returns dictionary of z-levels. The dictionary can be used either with integer or string key values: Integer
        keys return surface z-levels in ascending order (including domain boundary bottom and top). String keys
        (key = face_id) return the three z-levels of the face (metal bottom, metal-dielectric interface, dielectric top)

        The level z=0 is at lowest substrate top.
        """
        # Terms below z=0 level
        substrate_bottom = -float(self.ith_value(self.substrate_height, 0))
        z_levels = [substrate_bottom - self.lower_box_height] if self.lower_box_height > 0 else []
        z_levels += [substrate_bottom, 0.0]

        # Terms above z=0 level
        remaining_substrates = (len(self.face_stack) + 2 - len(z_levels)) // 2
        for s in range(remaining_substrates):
            z_levels.append(z_levels[-1] + float(self.ith_value(self.chip_distance, s)))
            z_levels.append(z_levels[-1] + float(self.ith_value(self.substrate_height, s + 1)))
        if len(z_levels) < len(self.face_stack) + 2:
            z_levels.append(z_levels[-1] + self.upper_box_height)

        # Create dictionary of z-levels including integer keys and face_id keys
        z_dict = dict(enumerate(z_levels))
        z_dict[-1] = z_levels[-1]
        for i, face_ids in enumerate(self.face_stack_list_of_lists()):
            sign = (-1) ** (i + int(self.lower_box_height > 0))
            base_z = z_levels[i + 1]
            metal_height = self.ith_value(self.metal_height, i)
            dielectric_height = self.ith_value(self.dielectric_height, i)
            for j, face_id in enumerate(face_ids):
                metal_z = base_z + sign * float(self.ith_value(metal_height, j))
                dielectric_z = metal_z + sign * float(self.ith_value(dielectric_height, j))
                z_dict[face_id] = [base_z, metal_z, dielectric_z]
                base_z = dielectric_z
        return z_dict

    def region_from_layer(self, face_id, layer_name):
        """ Returns a `Region` containing all geometry from a specified layer """
        face_layers = default_faces[face_id] if face_id in default_faces else dict()
        if layer_name in face_layers:
            return pya.Region(self.cell.begin_shapes_rec(self.layout.layer(face_layers[layer_name])))
        return pya.Region()

    def simplified_region(self, region, expansion=0.0):
        """ Returns a region that is simplified by functions region_with_merged_polygons and region_with_merged_points.
        More precisely:
        - Merges polygons ignoring gaps that are smaller than self.polygon_tolerance
        - Expands/shrinks region by amount given by 'expansion'
        - In each polygon of the region, removes points that are closer to other points than self.minimum_point_spacing
        """
        return region_with_merged_points(
            region_with_merged_polygons(region,
                                        tolerance=self.polygon_tolerance / self.layout.dbu,
                                        expansion=expansion / self.layout.dbu),
            tolerance=self.minimum_point_spacing / self.layout.dbu)

    def insert_layer(self, region, layer_name, **params):
        """Merges points in the `region` and inserts the result in a target layer. The params are forwarded to the
        'self.layers' dictionary.
        """
        if region.is_empty():
            return  # ignore the layer
        layer = get_simulation_layer_by_name(layer_name)
        self.cell.shapes(self.layout.layer(layer)).insert(region)
        if (pya.Region(self.box.to_itype(self.layout.dbu)) - region).is_empty():
            self.layers[layer_name] = params  # region cover self.box totally, so set layer without shape
        else:
            self.layers[layer_name] = {'layer': layer.layer, **params}

    def insert_stacked_up_layers(self, stack, z0):
        """Produces the layer stack-up and adds the layers into 'self.layers' dictionary.
        Each layer is split into sub-layers by their z-level.

        Args:
            stack: list of layers in form of tuples containing (region, layer name, thickness, material)
            z0: the base z-level for the layer stack-up
        """
        levels = dict()  # existing z-levels based on layers underneath (z-level as key and region as value)
        for region, layer_name, thickness, material in stack:
            if region.is_empty():
                continue

            # Split the layer into z-levels
            region_levels = dict()  # the layer region divided into z-levels
            non_region_levels = dict()  # the z-levels outside the layer region
            sum_reg = pya.Region()
            for z, reg in levels.items():
                intersection = reg & region
                if not intersection.is_empty():
                    region_levels[z] = intersection
                subtraction = reg - region
                if not subtraction.is_empty():
                    non_region_levels[z] = subtraction
                sum_reg += reg
            on_base_level = region - sum_reg
            if not on_base_level.is_empty():
                region_levels[round(z0, 12)] = on_base_level

            # Create collective layer for klayout visualization if the layer is split
            if len(region_levels) > 1:
                self.cell.shapes(self.layout.layer(get_simulation_layer_by_name(layer_name))).insert(region)

            # Apply parts of divided layers into self.layers
            for i, (z, reg) in enumerate(sorted(region_levels.items())):
                self.insert_layer(reg, f'{layer_name}_{i}' if len(region_levels) > 1 else layer_name,
                                  z=z, thickness=thickness, material=material)

            # Update existing z-levels dictionary
            if thickness != 0.0:
                levels = non_region_levels
                for z, reg in region_levels.items():
                    top_z = round(z + thickness, 12)
                    if top_z in levels:
                        levels[top_z] += reg
                    else:
                        levels[top_z] = reg

    def insert_layers_between_faces(self, i, opp_i, layer_name, **params):
        """Helper function to be used to produce indium bumps and TSVs"""
        z = self.face_z_levels()
        face_stack = self.face_stack_list_of_lists()
        box = self._face_box(i)
        if 0 <= opp_i < len(face_stack):
            box = box & self._face_box(opp_i)
        box_region = pya.Region(box.to_itype(self.layout.dbu))
        sum_region = pya.Region()
        for face_id in face_stack[i]:
            region = self.simplified_region(self.region_from_layer(face_id, layer_name) & box_region)
            if region.is_empty():
                continue
            sum_region += region
            if 0 <= opp_i < len(face_stack):
                for opp_id in face_stack[opp_i]:
                    common_region = region & self.simplified_region(
                        self.region_from_layer(opp_id, layer_name) & box_region)
                    if common_region.is_empty():
                        continue
                    if f'{opp_id}_{face_id}_{layer_name}' not in self.layers:  # if statement is to avoid duplicates
                        self.insert_layer(common_region, f'{face_id}_{opp_id}_{layer_name}', z=z[face_id][1],
                                          thickness=z[opp_id][1] - z[face_id][1], **params)
                    region -= common_region
                    if region.is_empty():
                        break
            if not region.is_empty():
                self.insert_layer(region, face_id + '_' + layer_name, z=z[face_id][1],
                                  thickness=z[opp_i + 1] - z[face_id][1], **params)
        return sum_region

    def create_simulation_layers(self):
        """Create the layers used for simulation export.

        Based on any geometry defined on the relevant lithography layers.

        This method is called from `__init__` after `build`, and should not be called directly.

        Geometry is added to layers created specifically for simulation purposes. The layer numbers, z-levels,
        thicknesses, materials, and other properties are stored in 'self.layers' parameter.

        In the simulation-specific layers, all geometry has been merged and converted to simple polygons, that is,
        polygons without holes.
        """

        z = self.face_z_levels()
        face_stack = self.face_stack_list_of_lists()
        for i, face_ids in enumerate(face_stack):
            sign = (-1) ** (i + int(self.lower_box_height > 0))
            stack = []
            dielectric_material = self.ith_value(self.dielectric_material, i)

            # insert TSVs and indium bumps
            tsv_params = {'edge_material': 'pec'} if self.hollow_tsv else {'material': 'pec'}
            tsv_region = self.insert_layers_between_faces(i, i - sign, "through_silicon_via", **tsv_params)
            bump_region = self.insert_layers_between_faces(i, i + sign, "indium_bump", material='pec')
            ground_box_region = pya.Region(self._face_box(i).to_itype(self.layout.dbu))

            for j, face_id in enumerate(face_ids):
                metal_gap_region = self.region_from_layer(face_id, "base_metal_gap_wo_grid")
                metal_add_region = self.region_from_layer(face_id, "base_metal_addition")
                lithography_region = self.simplified_region(metal_gap_region - metal_add_region, self.over_etching)

                if lithography_region.is_empty():
                    signal_region = pya.Region()
                    ground_region = ground_box_region.dup()
                else:
                    signal_region = ground_box_region - lithography_region

                    # Find the ground plane and subtract it from the simulation area
                    # First, add all polygons touching any of the edges
                    ground_region = pya.Region()
                    for edge in ground_box_region.edges():
                        ground_region += signal_region.interacting(edge)
                    # Now, remove all edge polygons which are also a port
                    if self.use_ports:
                        for port in self.ports:
                            if self.face_ids[port.face] == face_id:
                                if hasattr(port, 'ground_location'):
                                    v_mps = port.signal_location-port.ground_location
                                    v_mps = self.minimum_point_spacing * v_mps / v_mps.abs()
                                    signal_loc = (port.signal_location + v_mps).to_itype(self.layout.dbu)
                                    ground_region -= ground_region.interacting(pya.Edge(signal_loc, signal_loc))

                                    ground_loc = (port.ground_location - v_mps).to_itype(self.layout.dbu)
                                    ground_region += signal_region.interacting(pya.Edge(ground_loc, ground_loc))
                                else:
                                    signal_loc = port.signal_location.to_itype(self.layout.dbu)
                                    ground_region -= ground_region.interacting(pya.Edge(signal_loc, signal_loc))

                    ground_region.merge()
                    signal_region -= ground_region

                dielectric_region = ground_box_region - self.simplified_region(
                    self.region_from_layer(face_id, "dielectric_etch"))

                # Create gap and etch regions and update metals
                gap_region = ground_box_region - signal_region - ground_region  # excluding ground grid
                if self.with_grid:
                    ground_region -= self.ground_grid_region(face_id)
                etch_region = ground_box_region - signal_region - ground_region  # including ground grid
                signal_region -= tsv_region
                ground_region -= tsv_region
                dielectric_region -= tsv_region

                # Insert signal, ground and dielectric layers
                metal_thickness = z[face_id][1] - z[face_id][0]
                dielectric_thickness = z[face_id][2] - z[face_id][1]
                if self.fixed_level_stackup:
                    # Use fixed level stack-up
                    self.insert_layer(signal_region, face_id + "_signal", z=z[face_id][0], thickness=metal_thickness,
                                      material='pec')
                    self.insert_layer(ground_region, face_id + "_ground", z=z[face_id][0], thickness=metal_thickness,
                                      material='pec')
                    if dielectric_thickness != 0.0:
                        self.insert_layer(ground_box_region - dielectric_region, face_id + "_via", z=z[face_id][1],
                                          thickness=dielectric_thickness, material='pec')
                        subtract = [n for n in [face_id + "_signal", face_id + "_ground", face_id + "_via"]
                                    if n in self.layers and self.layers[n].get('thickness', 0.0) != 0.0]
                        self.insert_layer(ground_box_region, face_id + "_dielectric", z=z[face_id][0],
                                          thickness=z[face_id][2] - z[face_id][0],
                                          material=self.ith_value(dielectric_material, j),
                                          **({'subtract': subtract} if subtract else dict()))

                else:
                    # Use stack to produce drop-down stack-up
                    stack.append((signal_region, face_id + "_signal", metal_thickness, 'pec'))
                    stack.append((ground_region, face_id + "_ground", metal_thickness, 'pec'))
                    if dielectric_thickness != 0.0:
                        stack.append((dielectric_region, face_id + "_dielectric", dielectric_thickness,
                                      self.ith_value(dielectric_material, j)))

                # Insert gap and etch layers only on the first face of the stack-up (no material)
                if j == 0:
                    gap_z = z[face_id][0] - sign * self.vertical_over_etching
                    if self.vertical_over_etching > 0.0:
                        self.insert_layer(etch_region, face_id + "_etch", z=gap_z,
                                          thickness=sign * self.vertical_over_etching)
                    self.insert_layer(gap_region, face_id + "_gap", z=gap_z, thickness=z[face_id][1] - gap_z)

                # Insert airbridges
                bridge_z = sign * self.airbridge_height
                ab_flyover_region = self.simplified_region(self.region_from_layer(face_id, "airbridge_flyover")
                                                           ) & ground_box_region
                self.insert_layer(ab_flyover_region, face_id + "_airbridge_flyover", z=z[face_id][1] + bridge_z,
                                  thickness=0.0, material='pec')
                ab_pads_region = self.simplified_region(self.region_from_layer(face_id, "airbridge_pads")
                                                        ) & ground_box_region
                self.insert_layer(ab_pads_region, face_id + "_airbridge_pads", z=z[face_id][1], thickness=bridge_z,
                                  material='pec')

            self.insert_stacked_up_layers(stack, z[i + 1])

            # Rest of the features are not available with multilayer stack-up
            if len(face_ids) != 1:
                continue
            face_id = face_ids[0]

            # Create metal edge region
            metal_region = signal_region + ground_region
            me_region = pya.Region()
            n_terms = len(self.metal_edge_region_dimensions)
            if n_terms > 0:
                r_gap = float(self.metal_edge_region_dimensions[0])
                r_metal = float(self.metal_edge_region_dimensions[2 % n_terms])
                me_region = (metal_region.sized(r_gap / self.layout.dbu) & etch_region.sized(r_metal / self.layout.dbu)
                             & ground_box_region)

            # Insert TLS interface layers
            for layer_num, layer_id in enumerate(['MA', 'MS', 'SA']):
                layer_name = face_id + "_layer" + layer_id
                layer_z = [z[face_id][1], z[face_id][0], z[face_id][0] - sign * self.vertical_over_etching][layer_num]
                thickness = float(self.ith_value(self.tls_layer_thickness, layer_num))
                signed_thickness = [sign, -sign, -sign][layer_num] * thickness
                if self.tls_sheet_approximation:
                    params = {'z': layer_z + signed_thickness,
                              'thickness': 0.0}
                elif thickness != 0.0:
                    material = self.ith_value(self.tls_layer_material, layer_num)
                    params = {'z': layer_z,
                              'thickness': signed_thickness,
                              **(dict() if material is None else {'material': material})}

                    # Insert wall layer
                    wall_height = [z[face_id][0] - z[face_id][1], 0.0, sign * self.vertical_over_etching][layer_num]
                    if wall_height != 0.0:
                        wall_region = metal_region.sized(thickness / self.layout.dbu) & etch_region
                        self.insert_layer(wall_region, layer_name + "wall", z=layer_z, thickness=wall_height,
                                          **(dict() if material is None else {'material': material}))
                else:
                    continue

                # Insert layer
                layer_region = [metal_region.sized(thickness / self.layout.dbu) & (metal_region + etch_region -
                                                                                   bump_region - ab_pads_region),
                                metal_region, etch_region][layer_num]
                me_layer_region = me_region & layer_region
                if me_layer_region.is_empty():
                    self.insert_layer(layer_region, layer_name, **params)
                else:
                    self.insert_layer(me_layer_region, layer_name + "mer", **params)
                    self.insert_layer(layer_region - me_region, layer_name, **params)

            # Insert substrate and vacuum inside metal edge region
            if not me_region.is_empty():
                r_vacuum = float(self.metal_edge_region_dimensions[1 % n_terms])
                r_substrate = float(self.metal_edge_region_dimensions[3 % n_terms])

                if r_substrate != 0.0:
                    layers = ['_etch', '_through_silicon_via']
                    cond_layers = ['_layerMSmer', '_layerSAmer']
                    subtract = [k for k, v in self.layers.items() if face_id + '_' in k and
                                (any(k.endswith(t) for t in layers) or
                                 (any(k.endswith(t) for t in cond_layers) and v.get('material', None) is not None))]
                    self.insert_layer(me_region, face_id + "_substratemer", z=z[face_id][0] - sign * r_substrate,
                                      thickness=sign * r_substrate,
                                      material=self.ith_value(self.substrate_material,
                                                              (i + int(self.lower_box_height <= 0)) // 2),
                                      **({'subtract': subtract} if subtract else dict()))

                if r_vacuum + r_substrate != 0.0:
                    subtract = [n for n, v in self.layers.items() if face_id + '_' in n and
                                v.get('material', None) is not None and v.get('thickness', 0.0) != 0.0]
                    self.insert_layer(me_region, face_id + "_vacuummer", z=z[face_id][0] - sign * r_substrate,
                                      thickness=sign * (r_vacuum + r_substrate), material='vacuum',
                                      **({'subtract': subtract} if subtract else dict()))

        # Insert substrates
        for i in range(int(self.lower_box_height > 0), len(face_stack) + 1, 2):
            # faces around the substrate
            faces = []
            if i < len(face_stack):
                faces += face_stack[i]
            if i > 0:
                faces += face_stack[i-1]

            # find layers to be subtracted from substrate
            layers = ['_etch', '_through_silicon_via']
            cond_layers = ['_layerMS', '_layerSA', '_layerMSmer', '_layerSAmer', '_substratemer']
            subtract = [k for k, v in self.layers.items() if any(t + '_' in k for t in faces) and
                        (any(k.endswith(t) for t in layers) or
                         (any(k.endswith(t) for t in cond_layers) and v.get('material', None) is not None))]

            # insert substrate layer
            name = 'substrate' if len(face_stack) - int(self.lower_box_height > 0) < 2 else f'substrate_{i // 2}'
            self.insert_layer(pya.Region(self._face_box(i).to_itype(self.layout.dbu)), name, z=z[i],
                              thickness=z[i + 1] - z[i], material=self.ith_value(self.substrate_material, i // 2),
                              **({'subtract': subtract} if subtract else dict()))

        # Insert vacuum
        subtract = [n for n, v in self.layers.items() if v.get('material', None) is not None and
                    v.get('thickness', 0.0) != 0.0]
        self.insert_layer(pya.Region(self.box.to_itype(self.layout.dbu)), 'vacuum', z=z[0], thickness=z[-1] - z[0],
                          material='vacuum', **({'subtract': subtract} if subtract else dict()))

        # Eliminate gaps and overlaps caused by transformation to simple_polygon
        match_points_on_edges(self.cell, self.layout, [get_simulation_layer_by_name(n) for n in self.layers])

    def ground_grid_region(self, face_id):
        """Returns region of ground grid for the given face id."""
        grid_area = self.ground_grid_box * (1 / self.layout.dbu)
        protection = self.simplified_region(self.region_from_layer(face_id, "ground_grid_avoidance"))
        grid_mag_factor = 1
        return make_grid(grid_area, protection,
                         grid_step=10 * (1 / self.layout.dbu) * grid_mag_factor,
                         grid_size=5 * (1 / self.layout.dbu) * grid_mag_factor)

    def produce_waveguide_to_port(self, location, towards, port_nr, side=None,
                                  use_internal_ports=None, waveguide_length=None,
                                  term1=0, turn_radius=None,
                                  a=None, b=None, over_etching=None,
                                  airbridge=False, face=0):
        """Create a waveguide connection from some `location` to a port, and add the corresponding port to
        `simulation.ports`.

        Arguments:
            location (pya.DPoint): Point where the waveguide connects to the simulation
            towards (pya.DPoint): Point that sets the direction of the waveguide.
                The waveguide will start from `location` and go towards `towards`
            port_nr (int): Port index for the simulation engine starting from 1
            side (str): Indicate on which edge the port should be located. Ignored for internal ports.
                Must be one of `left`, `right`, `top` or `bottom`. If `None` then direction is inferred from `towards`.
            use_internal_ports (bool, optional): if True, ports will be inside the simulation. If False, ports will be
                brought out to an edge of the box, determined by `side`.
                Defaults to the value of the `use_internal_ports` parameter
            waveguide_length (float, optional): length of the waveguide (μm), used only for internal ports
                Defaults to the value of the `waveguide_length` parameter
            term1 (float, optional): Termination gap (μm) at `location`. Default 0
            turn_radius (float, optional): Turn radius of the waveguide. Not relevant for internal ports.
                Defaults to the value of the `r` parameter
            a (float, optional): Center conductor width. Defaults to the value of the `a` parameter
            b (float, optional): Conductor gap width. Defaults to the value of the `b` parameter
            over_etching (float, optional): Expansion of gaps. Defaults to the value of the `over_etching` parameter
            airbridge (bool, optional): if True, an airbridge will be inserted at `location`. Default False.
            face: face to place waveguide and port on. Either 0 (default) or 1, for bottom or top face.
        """

        waveguide_safety_overlap = 0.005  # Extend waveguide by this amount to avoid gaps due to nm-scale rounding
                                          # errors
        waveguide_gap_extension = 1  # Extend gaps beyond waveguides into ground plane to define the ground port edge

        if turn_radius is None:
            turn_radius = self.r
        if a is None:
            a = self.a
        if b is None:
            b = self.b
        if over_etching is None:
            over_etching = self.over_etching
        if use_internal_ports is None:
            use_internal_ports = self.use_internal_ports
        if waveguide_length is None:
            waveguide_length = self.waveguide_length

        # Create a new path in the direction of path but with length waveguide_length
        direction = towards - location
        direction = direction / direction.length()

        waveguide_a = a
        waveguide_b = b
        # First node may be an airbridge
        if airbridge:
            first_node = Node(location, Airbridge, a=a, b=b)
            waveguide_a = Airbridge.bridge_width
            waveguide_b = Airbridge.bridge_width / Element.a * Element.b
        else:
            first_node = Node(location)

        if use_internal_ports:
            signal_point = location + (waveguide_length + over_etching) * direction
            ground_point = location + (waveguide_length + a - 3 * over_etching) * direction

            nodes = [
                first_node,
                Node(signal_point + waveguide_safety_overlap * direction),
            ]
            port = InternalPort(port_nr, *self.etched_line(signal_point, ground_point), face=face)

            extension_nodes = [
                Node(ground_point),
                Node(ground_point + waveguide_gap_extension * direction)
            ]
        else:
            corner_point = location + (waveguide_length + turn_radius) * direction
            if side is None:  # infer EdgePort location if not given
                d = towards - location
                side = (("left", "right"), ("bottom", "top"))[abs(d.x) < abs(d.y)][d.x + d.y > 0]
            port_edge_point = {
                "left": pya.DPoint(self.box.left, corner_point.y),
                "right": pya.DPoint(self.box.right, corner_point.y),
                "top": pya.DPoint(corner_point.x, self.box.top),
                "bottom": pya.DPoint(corner_point.x, self.box.bottom)
            }[side]

            nodes = [
                first_node,
                Node(corner_point),
                Node(port_edge_point),
            ]
            port = EdgePort(port_nr, port_edge_point, face=face)

        tl = self.add_element(WaveguideComposite,
                              nodes=nodes,
                              r=turn_radius,
                              term1=term1,
                              term2=0,
                              a=waveguide_a,
                              b=waveguide_b,
                              face_ids=[self.face_ids[face]]
                              )

        self.cell.insert(pya.DCellInstArray(tl.cell_index(), pya.DTrans()))
        feedline_length = tl.length()

        if use_internal_ports:
            port_end_piece = self.add_element(WaveguideComposite,
                                              nodes=extension_nodes,
                                              a=a,
                                              b=b,
                                              term1=a-4*over_etching,
                                              term2=0,
                                              face_ids=[self.face_ids[face]]
                                              )
            self.cell.insert(pya.DCellInstArray(port_end_piece.cell_index(), pya.DTrans()))
        else:
            port.deembed_len = feedline_length

        self.ports.append(port)

    def get_parameters(self):
        """Return dictionary with all parameters and their values."""

        return {param: getattr(self, param) for param in type(self).get_schema()}

    def etched_line(self, p1: pya.DPoint, p2: pya.DPoint):
        """
            Return the end points of line segment after extending it at both ends by amount of over_etching.
            This function must be used when initializing InternalPort.

            Arguments:
                p1 (pya.DPoint): first end of line segment
                p2 (pya.DPoint): second end of line segment

            Returns:
                 [p1 - d, p2 + d]: list of extended end points.
        """
        d = (self.over_etching / p1.distance(p2)) * (p2 - p1)
        return [p1 - d, p2 + d]

    def get_port_data(self):
        """ Return the port data in dictionary form and add the information of port polygon. Includes following:

            * Items from `Port` instance
            * polygon: point coordinates of the port polygon
            * signal_edge: point coordinates of the signal edge
            * ground_edge: point coordinates of the ground edge
        """
        simulation = self
        z = self.face_z_levels()
        # gather port data
        port_data = []
        if simulation.use_ports:
            for port in simulation.ports:
                # Basic data from Port
                p_data = port.as_dict()

                face_id = self.face_ids[port.face]

                # Define a 3D polygon for each port
                if isinstance(port, EdgePort):

                    ps = (simulation.port_size if isinstance(simulation.port_size, list)
                          else [simulation.port_size / 2] * 4)

                    port_z0 = max(z[face_id][0] - ps[2], z[0])
                    port_z1 = min(z[face_id][0] + ps[3], z[-1])

                    # Determine which edge this port is on
                    port_x0 = port_x1 = port.signal_location.x
                    port_y0 = port_y1 = port.signal_location.y
                    if abs(port.signal_location.x - simulation.box.left) < self.layout.dbu:
                        port_y0 = max(port.signal_location.y - ps[1], simulation.box.bottom)
                        port_y1 = min(port.signal_location.y + ps[0], simulation.box.top)
                    elif abs(port.signal_location.x - simulation.box.right) < self.layout.dbu:
                        port_y0 = max(port.signal_location.y - ps[0], simulation.box.bottom)
                        port_y1 = min(port.signal_location.y + ps[1], simulation.box.top)
                    elif abs(port.signal_location.y - simulation.box.bottom) < self.layout.dbu:
                        port_x0 = max(port.signal_location.x - ps[0], simulation.box.left)
                        port_x1 = min(port.signal_location.x + ps[1], simulation.box.right)
                    elif abs(port.signal_location.y - simulation.box.top) < self.layout.dbu:
                        port_x0 = max(port.signal_location.x - ps[1], simulation.box.left)
                        port_x1 = min(port.signal_location.x + ps[0], simulation.box.right)
                    else:
                        raise ValueError(f"Port {port.number} is an EdgePort but not on the edge of the simulation box")

                    p_data['polygon'] = [
                        [port_x0, port_y0, port_z0],
                        [port_x1, port_y1, port_z0],
                        [port_x1, port_y1, port_z1],
                        [port_x0, port_y0, port_z1]
                    ]

                elif isinstance(port, InternalPort):
                    if hasattr(port, 'ground_location'):
                        try:
                            _, _, signal_edge = find_edge_from_point_in_cell(
                                simulation.cell,
                                # simulation.get_layer(port.signal_layer, port.face),
                                self.layout.layer(get_simulation_layer_by_name(face_id + '_' + port.signal_layer)),
                                port.signal_location,
                                simulation.layout.dbu)
                            _, _, ground_edge = find_edge_from_point_in_cell(
                                simulation.cell,
                                # simulation.get_layer('simulation_ground', port.face),
                                self.layout.layer(get_simulation_layer_by_name(face_id + '_ground')),
                                port.ground_location,
                                simulation.layout.dbu)

                            port_z = z[face_id][0]
                            p_data['polygon'] = get_enclosing_polygon(
                                [[signal_edge.x1, signal_edge.y1, port_z], [signal_edge.x2, signal_edge.y2, port_z],
                                 [ground_edge.x1, ground_edge.y1, port_z], [ground_edge.x2, ground_edge.y2, port_z]])
                            p_data['signal_edge'] = ((signal_edge.x1, signal_edge.y1, port_z),
                                                     (signal_edge.x2, signal_edge.y2, port_z))
                            p_data['ground_edge'] = ((ground_edge.x1, ground_edge.y1, port_z),
                                                     (ground_edge.x2, ground_edge.y2, port_z))
                        except ValueError as e:
                            logging.warning('Unable to create polygon for port %s, because either signal or ground '
                                                'edge is not found.', port.number)
                            logging.debug(e)
                else:
                    raise ValueError("Port {} has unsupported port class {}".format(port.number, type(port).__name__))

                # Change signal and ground location from DVector to list and add z-component as third term
                for location in ['signal_location', 'ground_location']:
                    if location in p_data:
                        p_data[location] = [p_data[location].x, p_data[location].y, z[face_id][0]]

                port_data.append(p_data)

        return port_data

    def get_simulation_data(self):
        """ Return the simulation data in dictionary form. Contains following:

            * gds_file: name of gds file to include geometry layers,
            * units: length unit in simulations, 'um',
            * layers: geometry data,
            * material_dict: Dictionary of dielectric materials,
            * box: Boundary box,
            * ports: Port data in dictionary form, see self.get_port_data(),
            * parameters: All Simulation class parameters in dictionary form,
        """
        # check that materials are defined in material_dict
        materials = [layer['material'] for layer in self.layers.values() if 'material' in layer]
        mater_dict = ast.literal_eval(self.material_dict) if isinstance(self.material_dict, str) else self.material_dict
        for name in materials:
            if name not in mater_dict and name not in ['pec', 'vacuum', None]:
                raise ValueError("Material '{}' used but not defined in Simulation.material_dict".format(name))

        return {
            'gds_file': self.name + '.gds',
            'units': 'um',  # hardcoded assumption in multiple places
            'layers': self.layers,
            'material_dict': mater_dict,
            'box': self.box,
            'ports': self.get_port_data(),
            'parameters': self.get_parameters(),
        }

    def get_layers(self):
        """ Returns simulation layer numbers in list. Only return layers that are in use. """
        return [pya.LayerInfo(d['layer'], 0) for d in self.layers.values() if 'layer' in d]

    @staticmethod
    def delete_instances(cell, name, index=(0,)):
        """
            Allows for deleting a sub-cell of the top 'cell' with a specific 'name'. The 'index' argument can be used to
            access more than one sub-cell sharing the same 'name', but with different appended 'index' to the 'name'.
        """
        for i in index:
            index_name = f'${i}' if i > 0 else ''
            cell_to_be_deleted = cell.layout().cell(f'{name}{index_name}')
            if cell_to_be_deleted is not None:
                # This has started causing errors on KLayout's side, protect with try-except
                try:
                    cell_to_be_deleted.delete()
                except RuntimeError as e:
                    logging.warning(f"Attempt to delete cell {name}{index_name} caused following error: {e}")



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import numpy

from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt

from kqcircuits.chips.chip import Chip


class Stripes(Chip):
    """The PCell declaration for a Stripes chip."""

    edge_len = Param(pdt.TypeInt, "Length of square's one edge", 80)
    inter_space = Param(pdt.TypeInt, "Space in between the Squares", 20)
    axis = Param(pdt.TypeString, "The axis of the stripes", "Vertical")

    def build(self):

        # defining the dimensions for creating the polygonal area of test
        left = self.box.left
        right = self.box.right
        top = self.box.top
        bottom = self.box.bottom

        # dimensions for the hole on the ground plane
        p_top = 8000
        p_bottom = 2000
        p_left = 2000
        p_right = 8000

        width = right - left
        height = top - bottom
        dice_width = float(self.frames_dice_width[0])

        # create the test area polygon
        poly = pya.DPolygon([
            pya.DPoint(left + p_left, bottom + 100 + dice_width),
            pya.DPoint(left + p_left, bottom + p_bottom),
            pya.DPoint(left + 100 + dice_width, bottom + p_bottom),
            pya.DPoint(left + 100 + dice_width, bottom + p_top),
            pya.DPoint(left + p_left, bottom + p_top),
            pya.DPoint(left + p_left, bottom + height - 100 - dice_width),
            pya.DPoint(left + p_right, bottom + height - 100 - dice_width),
            pya.DPoint(left + p_right, bottom + p_top),
            pya.DPoint(left + width - 100 - dice_width, bottom + p_top),
            pya.DPoint(left + width - 100 - dice_width, bottom + p_bottom),
            pya.DPoint(left + p_right, bottom + p_bottom),
            pya.DPoint(left + p_right, bottom + 100 + dice_width)
        ])

        # create the box array
        b_array = []

        square_y_reach = bottom + height
        square_x_start = left + 120
        square_x_reach = left + width - dice_width

        step = self.edge_len + self.inter_space
        stripe_step = step * 2

        if self.axis == "Vertical":
            stripe_start = left + dice_width + self.inter_space
            stripe_reach = left + width

            stripe_top_right_y = top - 20
            stripe_bottom_left_y = bottom + 20

            square_y_start = bottom + dice_width + self.inter_space

            square_y_step = step
            square_x_step = step * 2

        elif self.axis == "Horizontal":
            stripe_start = bottom + dice_width + self.inter_space
            stripe_reach = bottom + height

            stripe_top_right_x = right - 20
            stripe_bottom_left_x = left + 20

            square_y_start = bottom - 80

            square_y_step = step * 2
            square_x_step = step

        # creating the stripes for all of the ground grid
        for c in numpy.arange(stripe_start, stripe_reach, stripe_step):
            if self.axis == "Vertical":
                stripe_top_right_x = c + self.edge_len
                stripe_bottom_left_x = c

            elif self.axis == "Horizontal":
                stripe_top_right_y = c + self.edge_len
                stripe_bottom_left_y = c

            r = pya.DPolygon(
                pya.DBox(stripe_bottom_left_x, stripe_bottom_left_y, stripe_top_right_x, stripe_top_right_y)).to_itype(
                self.layout.dbu)
            b_array.append(r)

        # creating the squares for all of the ground grid
        for y in numpy.arange(square_y_start, square_y_reach, square_y_step):
            for x in numpy.arange(square_x_start, square_x_reach, square_x_step):
                b = pya.DPolygon(pya.DBox(x, y, x + self.edge_len, y + self.edge_len)).to_itype(self.layout.dbu)
                b_array.append(b)

        # substract the box array from the polygon
        reg1 = pya.Region(poly.to_itype(self.layout.dbu))
        reg2 = pya.Region()

        for b in b_array:
            reg2.insert(b)

        result = reg1 - reg2
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(result)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import os
from importlib.util import module_from_spec, spec_from_file_location
from setuptools import setup, find_packages


# Loads _version.py module without importing the whole package.
def get_version_and_cmdclass(package_name):
    spec = spec_from_file_location('version', os.path.join(package_name, '_version.py'))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.cmdclass


version, cmdclass = get_version_and_cmdclass('kqcircuits')

setup(
    name='kqcircuits',
    version=version,
    cmdclass=cmdclass,
    description="KQCircuits is a KLayout/Python-based superconducting quantum circuit library developed by IQM.",
    long_description=open('README.md').read(),  # pylint: disable=consider-using-with
    long_description_content_type='text/markdown',
    author="IQM Finland Oy",
    author_email="kqcircuits@meetiqm.com",
    url="https://iqm-finland.github.io/KQCircuits/",
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.6.9",
    install_requires=[                # Record dependencies in kqcircuits/util/dependencies.py too
        "klayout>=0.28",
        "numpy>=1.16",
        "scipy>=1.2",
        "tqdm>=4.61",
        # psutil was considered when cpu_count(logical=False), was implemented in an alternative way
        # in elmer_export.py and gmsh_helpers.py, consider adding if more features are needed.
    ],
    extras_require={
        "docs": ["sphinx~=4.4", "sphinx-rtd-theme~=0.4", "networkx>=2.7,<3.2", "matplotlib>=3.5.1"],
        "tests": ["pytest>=6.0.2", "pytest-cov~=2.8", "pytest-xdist>=2.1", "tox>=3.18", "pylint==2.9",
                  "networkx>=2.7,<3.2", "matplotlib>=3.5.1"],
        "notebooks": ["jupyter~=1.0.0", "klayout>=0.28"],
        "graphs": ["networkx>=3.0,<3.2", "matplotlib>=3.6.3"],
        "simulations": ["gmsh>=4.11.1", "pandas>=1.5.3"],
    },
    entry_points={
        'console_scripts':[
            'kqc = console_scripts.run:run',
            ]
        }
)



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from math import pi
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.element import Element
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.util.refpoints import WaveguideToSimPort


class HangerResonator(Element):
    """
    Hanger Resonator
    """
    coupling_length = Param(pdt.TypeDouble, "Length of the resonator center part (coupling)", 500, unit="μm")
    head_length = Param(pdt.TypeDouble, "Length of the resonator left waveguide (head) ", 300, unit="μm")
    resonator_length = Param(pdt.TypeDouble, "Total length of the resonator", 1000, unit="μm")

    res_a = Param(pdt.TypeDouble, "Trace width of resonator line", 10, unit="μm")
    res_b = Param(pdt.TypeDouble, "Gap width of resonator line", 6, unit="μm")

    ground_width = Param(pdt.TypeDouble, "Trace width of middle ground", 10, unit="μm")

    def build(self):

        # If turn radius is smaller than half of the trace width it will create some overlapping masks and sharp angles
        if self.r < self.res_a / 2:
            self.raise_error_on_cell(f'Turn radius must be at least res_a/2, now r={self.r}, res_a/2={self.res_a/2}')

        # probe line, origin at the center of the trace
        points_pl = [pya.DPoint(0, 0)]
        points_pl.append(pya.DPoint(self.coupling_length, 0))

        # distance from origin to start of the wg trace
        wg_start_height = -self.a / 2 - self.b - self.ground_width - self.res_b
        # x distance from pl port to center trace of vertical waveguides
        corner_x = self.r
        # corner arc length
        corner_length = pi * self.r/2
        head_length_down = self.head_length - corner_length

        points = []

        if head_length_down > 0:
            # left side, head
            # left leg
            p1 = pya.DPoint(-corner_x, wg_start_height - self.res_a / 2 - self.r - head_length_down)
            points.append(p1)
            # corner
            p2 = pya.DPoint(-corner_x, wg_start_height - self.res_a / 2)

            length_without_tail = self.head_length + self.coupling_length + corner_length

        # If head lenght is too small don't create the curve on left side
        else:
            # Add a stub corresponding to head length if head length is shorter than the corner
            p2 = pya.DPoint(-max(self.head_length, 0), wg_start_height - self.res_a/2)

            length_without_tail = self.coupling_length + corner_length

        points.append(p2)


        # If given resonator length is too small, don't create downwards tail
        if length_without_tail >= self.resonator_length:

            # Add a stub corresponding to tail length if head length is shorter than the corner
            x_tail_offset = max(0, self.resonator_length - (length_without_tail - corner_length))

            p3 = pya.DPoint(self.coupling_length + x_tail_offset, wg_start_height - self.res_a / 2)

            points.append(p3)

        else:
            tail_length = self.resonator_length - length_without_tail
            # right leg (tail)
            p3 = pya.DPoint(self.coupling_length + corner_x, wg_start_height - self.res_a / 2)
            p4 = pya.DPoint(self.coupling_length + corner_x, wg_start_height - self.res_a / 2 - self.r - tail_length)

            points.append(p3)
            points.append(p4)

        cells_pl, _ = self.insert_cell(WaveguideCoplanar, path=points_pl)
        cells_resonator, _ = self.insert_cell(WaveguideCoplanar, path=points, a=self.res_a, b=self.res_b)

        self.copy_port("a", cells_pl)
        self.copy_port("b", cells_pl)
        self.copy_port("a", cells_resonator, "resonator_a")
        self.copy_port("b", cells_resonator, "resonator_b")
        # these are needed for simulations
        self.copy_port("a", cells_pl, "sim_a")
        self.copy_port("b", cells_pl, "sim_b")

    @classmethod
    def get_sim_ports(cls, simulation):
        return [WaveguideToSimPort("port_sim_a", use_internal_ports=False, a=simulation.a, b=simulation.b),
                WaveguideToSimPort("port_sim_b", use_internal_ports=False, a=simulation.a, b=simulation.b),
                WaveguideToSimPort("port_resonator_a", a=simulation.res_a, b=simulation.res_b),
                WaveguideToSimPort("port_resonator_b", a=simulation.res_a, b=simulation.res_b)]



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.simulations.simulation import Simulation
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.chips.xmons_direct_coupling import XMonsDirectCoupling
from kqcircuits.simulations.port import InternalPort
from kqcircuits.elements.fluxlines.fluxline import Fluxline
from kqcircuits.junctions.junction import Junction


@add_parameters_from(Junction, junction_type="Sim")
@add_parameters_from(Fluxline, fluxline_type="none")
@add_parameters_from(XMonsDirectCoupling, "arm_width_a", "rr_cpl_width")
class XMonsDirectCouplingSim(Simulation):

    #Re-define some XMonsDirectCoupling parameters with defaults changed:
    qubit_spacing = Param(pdt.TypeDouble, "Qubit spacing", 3, unit="μm")
    arm_width_b = Param(pdt.TypeDouble, "Qubit 2 arm width", 66, unit="μm")
    waveguide_length = Param(pdt.TypeDouble,
                             "Length of waveguide stubs or distance between couplers and waveguide turning point", 100)
    cpl_width = Param(pdt.TypeDouble, "Qubit RR coupler width", 24, unit="μm")
    junction_inductances = Param(pdt.TypeList, "Qubit junction inductances",
                                [13.5e-9, 13.5e-9, 13.5e-9], unit="[H, H, H]")
    junction_capacitances = Param(pdt.TypeList, "Qubit junction capacitances",
                                [0.1e-15, 0.1e-15, 0.1e-15], unit="[F, F, F]")

    produce_qubits = XMonsDirectCoupling.produce_qubits

    def build(self):

        # parameter used by produce_qubits
        self.rr_cpl_width = (self.cpl_width, self.cpl_width, self.cpl_width)

        self.produce_qubits()

        # driveline 1
        self.produce_waveguide_to_port(
            port_nr=1,
            location=self.refpoints["QB1_port_drive"],
            term1=10,
            towards=self.refpoints["QB1_port_drive"] + pya.DVector(-1, 1)
        )
        # driveline 2
        self.produce_waveguide_to_port(
            port_nr=2,
            location=self.refpoints["QB2_port_drive"],
            term1=10,
            towards=self.refpoints["QB2_port_drive"] + pya.DVector(0, -1)
        )
        # driveline 3
        self.produce_waveguide_to_port(
            port_nr=3,
            location=self.refpoints["QB3_port_drive"],
            term1=10,
            towards=self.refpoints["QB3_port_drive"] + pya.DVector(1, 1),
        )

        # readout 1
        self.produce_waveguide_to_port(
            port_nr=4,
            a=5,
            b=20,
            location=self.refpoints["QB1_port_cplr1"],
            towards=self.refpoints["QB1_port_cplr1"] + pya.DVector(0, 1),
            use_internal_ports=True
        )
        # readout 2
        self.produce_waveguide_to_port(
            port_nr=5,
            a=5,
            b=20,
            location=self.refpoints["QB2_port_cplr1"],
            towards=self.refpoints["QB2_port_cplr1"] + pya.DVector(0, 1),
            use_internal_ports=True
        )
        # readout 3
        self.produce_waveguide_to_port(
            port_nr=6,
            a=5,
            b=20,
            location=self.refpoints["QB3_port_cplr1"],
            towards=self.refpoints["QB3_port_cplr1"] + pya.DVector(0, 1),
            use_internal_ports=True
        )

        # qubits
        self.ports.extend([
            InternalPort(7, *self.etched_line(self.refpoints["QB1_port_squid_a"], self.refpoints["QB1_port_squid_b"]),
                inductance=self.junction_inductances[0], capacitance=self.junction_capacitances[0], junction=True),
            InternalPort(8, *self.etched_line(self.refpoints["QB2_port_squid_a"], self.refpoints["QB2_port_squid_b"]),
                inductance=self.junction_inductances[1], capacitance=self.junction_capacitances[1], junction=True),
            InternalPort(9, *self.etched_line(self.refpoints["QB3_port_squid_a"], self.refpoints["QB3_port_squid_b"]),
                inductance=self.junction_inductances[2], capacitance=self.junction_capacitances[2], junction=True),
        ])



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.elements.element import Element
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt


class LauncherDC(Element):
    """The PCell declaration for a DC launcher for connecting wirebonds.

    .. MARKERS_FOR_PNG 0,0 80,0 -278,-70
    """

    width = Param(pdt.TypeDouble, "Pad width", 500, unit="μm")

    def build(self):

        extra_width = 100

        offset = self.width/2
        metal_region = pya.Region((pya.DBox(-offset, -offset, offset, offset)).to_itype(self.layout.dbu))

        offset = (self.width + extra_width)/2
        gap_region = pya.Region((pya.DBox(-offset, -offset, offset, offset)).to_itype(self.layout.dbu))
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(gap_region - metal_region)

        offset = (self.width + extra_width)/2 + self.margin
        shape = pya.Region((pya.DBox(-offset, -offset, offset, offset)).to_itype(self.layout.dbu))
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(shape)

        # add reference point
        self.add_port("", pya.DPoint(0, 0))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import math
from kqcircuits.pya_resolver import pya
from kqcircuits.util.label import produce_label, LabelOrigin
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.element import Element
from kqcircuits.elements.markers.marker import Marker
from kqcircuits.defaults import default_brand, default_marker_type, default_chip_label_face_prefixes


class ChipFrame(Element):
    """The PCell declaration for a chip frame.

    The chip frame consists of a dicing edge, and labels and markers in the corners.

     .. MARKERS_FOR_PNG 0,5000,10000,5000 5000,0,5000,10000
    """

    box = Param(pdt.TypeShape, "Border", pya.DBox(pya.DPoint(0, 0), pya.DPoint(10000, 10000)),
                docstring="Bounding box of the chip frame")
    dice_width = Param(pdt.TypeDouble, "Dicing width", 200, unit="μm")
    dice_grid_margin = Param(pdt.TypeDouble, "Margin between dicing edge and ground grid", 100,
                             docstring="Margin of the ground grid avoidance layer for dicing edge")
    name_mask = Param(pdt.TypeString, "Name of the mask", "M000")
    name_chip = Param(pdt.TypeString, "Name of the chip", "CTest")
    name_copy = Param(pdt.TypeString, "Name of the copy", None)
    name_brand = Param(pdt.TypeString, "Name of the brand", default_brand)
    text_margin = Param(pdt.TypeDouble, "Margin for labels", 100,
                        docstring="Margin of the ground grid avoidance layer around the text")
    marker_dist = Param(pdt.TypeDouble, "Marker distance from edges", 1500,
                        docstring="Distance of markers from closest edges of the chip face")
    diagonal_squares = Param(pdt.TypeInt, "Number of diagonal squares for the markers", 10)
    use_face_prefix = Param(pdt.TypeBoolean, "Use face prefix for chip name label", False)
    marker_types = Param(pdt.TypeList, "Marker type for each chip corner, clockwise starting from lower left",
                         default=[default_marker_type] * 4)
    chip_dicing_width = Param(pdt.TypeDouble, "Width of the chip dicing reference line", 10.0, unit="µm")
    chip_dicing_line_length = Param(pdt.TypeDouble, "Length of the chip dicing reference line", 100.0, unit="µm")
    chip_dicing_gap_length = Param(pdt.TypeDouble, "Gap between two chip dicing reference dashes", 50.0, unit="µm")
    chip_dicing_in_base_metal = Param(pdt.TypeBoolean, "Insert chip dicing lines in base metal addition", False)

    def build(self):
        """Produces dicing edge, markers, labels and ground grid for the chip face."""
        self._produce_dicing_edge()
        self._produce_labels()
        self._produce_markers()

    def _produce_labels(self):
        x_min, x_max, y_min, y_max = self._box_points()
        if self.use_face_prefix:
            face_id = self.face()["id"]
            face_prefix = default_chip_label_face_prefixes[face_id].upper() \
                if default_chip_label_face_prefixes and (face_id in default_chip_label_face_prefixes) \
                else face_id.upper()
            chip_name = face_prefix + self.name_chip
        else:
            chip_name = self.name_chip
        labels = [self.name_mask, chip_name, self.name_copy, self.name_brand]
        if self.name_mask:
            self._produce_label(labels[0], pya.DPoint(x_min, y_max), LabelOrigin.TOPLEFT)
        if self.name_chip:
            self._produce_label(labels[1], pya.DPoint(x_max, y_max), LabelOrigin.TOPRIGHT)
        self._produce_label(labels[2], pya.DPoint(x_max, y_min), LabelOrigin.BOTTOMRIGHT)
        self._produce_label(labels[3], pya.DPoint(x_min, y_min), LabelOrigin.BOTTOMLEFT)

    def _produce_label(self, label, location, origin):
        """Produces Text PCells with text `label` with `origin` of the text at `location`.

        Wrapper for the stand alone function `produce_label`.
        Text size scales with chip dimension for chips smaller than 7 mm.

        Args:
            label: the produced text
            location: DPoint of the location of the text
            origin: LabelOrigin of the corner of the label to be placed at the location

        Effect:
            label PCells added to the layout into the parent PCell
        """
        size = 350 * min(1, self.box.width() / 7000, self.box.height() / 7000)
        produce_label(self.cell, label, location, origin, self.dice_width, self.text_margin,
                      [self.face()["base_metal_gap_wo_grid"], self.face()["base_metal_gap_for_EBL"]],
                      self.face()["ground_grid_avoidance"], size)

    def _produce_markers(self):
        x_min, x_max, y_min, y_max = self._box_points()
        if len(self.marker_types) == 4:
            self._produce_marker(self.marker_types[0], pya.DTrans(x_min + self.marker_dist, y_min + self.marker_dist) \
                                 * pya.DTrans.R180, self.face()["id"] + "_marker_sw")
            self._produce_marker(self.marker_types[3], pya.DTrans(x_max - self.marker_dist, y_min + self.marker_dist) \
                                 * pya.DTrans.R270, self.face()["id"] + "_marker_se")
            self._produce_marker(self.marker_types[1], pya.DTrans(x_min + self.marker_dist, y_max - self.marker_dist) \
                                 * pya.DTrans.R90, self.face()["id"] + "_marker_nw")
            self._produce_marker(self.marker_types[2], pya.DTrans(x_max - self.marker_dist, y_max - self.marker_dist) \
                                 * pya.DTrans.R0, self.face()["id"] + "_marker_ne")
        else:
            print('Warning: chip frame markers need to be for all four corners')

    def _produce_marker(self, marker_type, trans, name):
        if not marker_type:
            return
        cell_marker = self.add_element(Marker, marker_type=marker_type)
        self.insert_cell(cell_marker, trans)
        self.refpoints[name] = trans.disp

    def _produce_dicing_edge(self):
        shape = pya.DPolygon(self._border_points(self.dice_width))
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(shape)
        self.cell.shapes(self.get_layer("base_metal_gap_for_EBL")).insert(shape)

        protection = pya.DPolygon(self._border_points(self.dice_width + self.dice_grid_margin, self.margin))
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(protection)

        box_points = self._box_points()
        p1 = pya.DPoint(box_points[0], box_points[2])
        p2 = pya.DPoint(box_points[1], box_points[2])
        p3 = pya.DPoint(box_points[0], box_points[3])
        p4 = pya.DPoint(box_points[1], box_points[3])
        self._produce_lines_along_edge(p1.y, p3.y, True, p1)
        self._produce_lines_along_edge(p1.x, p2.x, False, p1)
        self._produce_lines_along_edge(p2.y, p4.y, True, p2)
        self._produce_lines_along_edge(p3.x, p4.x, False, p3)

    def _box_points(self):
        """Returns x_min, x_max, y_min, y_max for the given box."""
        x_min = min(self.box.p1.x, self.box.p2.x)
        x_max = max(self.box.p1.x, self.box.p2.x)
        y_min = min(self.box.p1.y, self.box.p2.y)
        y_max = max(self.box.p1.y, self.box.p2.y)
        return x_min, x_max, y_min, y_max

    def _border_points(self, w, extension=0):
        """Returns a set of points forming frame with outer edge on the chip boundaries, and frame thickness ``w``.
        Optional parameter ``extension`` extends the outer edge of the box.
        """
        x_min, x_max, y_min, y_max = self._box_points()
        points = [
            pya.DPoint(x_min - extension, y_min - extension),
            pya.DPoint(x_max + extension, y_min - extension),
            pya.DPoint(x_max + extension, y_max + extension),
            pya.DPoint(x_min - extension, y_max + extension),
            pya.DPoint(x_min - extension, y_min - extension),

            pya.DPoint(x_min + w, y_min + w),
            pya.DPoint(x_min + w, y_max - w),
            pya.DPoint(x_max - w, y_max - w),
            pya.DPoint(x_max - w, y_min + w),
            pya.DPoint(x_min + w, y_min + w),
        ]
        return points

    def _produce_lines_along_edge(self, line_start, line_end, is_vertical, position):
        """Adds chip dicing reference lines along one edge of the chip.
        The line pattern is tied to the global coordinate system

        Args:
            line_start: One-dimensional coordinate of start of the line
            line_end: One-dimensional coordinate of end of the line
            is_vertical: True to draw a line along y-axis, False to draw along x-axis
            position: A DPoint that is in the line
        """
        start = line_start
        segment_length = self.chip_dicing_line_length + self.chip_dicing_gap_length
        end = math.floor(line_start / segment_length) * segment_length + self.chip_dicing_line_length
        if end > start:
            self._add_chip_dicing_line_dash(start, end, is_vertical, position)
        while True:
            start = end + self.chip_dicing_gap_length
            if start > line_end:
                break
            end = start + self.chip_dicing_line_length
            if end > line_end:
                self._add_chip_dicing_line_dash(start, line_end, is_vertical, position)
                break
            self._add_chip_dicing_line_dash(start, end, is_vertical, position)

    def _add_chip_dicing_line_dash(self, start, end, is_vertical, position):
        """Adds a dash as part of a chip dicing reference line.

        Args:
            start: One-dimensional coordinate of start of the dash
            end: One-dimensional coordinate of end of the dash
            is_vertical: True to draw a dash along y-axis, False to draw along x-axis
            position: A DPoint that is in the line to which the dash belongs to
        """
        if is_vertical:
            box = pya.DBox(position.x - self.chip_dicing_width / 2, start, position.x + self.chip_dicing_width / 2, end)
        else:
            box = pya.DBox(start, position.y - self.chip_dicing_width / 2, end, position.y + self.chip_dicing_width / 2)
        self.cell.shapes(self.get_layer("chip_dicing")).insert(box)
        if self.chip_dicing_in_base_metal:
            self.cell.shapes(self.get_layer("base_metal_addition")).insert(box)



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.elements.airbridges.airbridge import Airbridge
from kqcircuits.elements.flip_chip_connectors.flip_chip_connector_dc import FlipChipConnectorDc
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from


@add_parameters_from(FlipChipConnectorDc)
@add_parameters_from(Airbridge, bridge_width=40, pad_length=40)
class AirbridgeMultiFace(Airbridge):
    """PCell declaration for a multi-face equivalent for an airbridge.

    Origin is at the geometric center. The airbridge is in vertical direction.

    Adds base metal on the second face and optionally creates bumps at both ends.

    .. MARKERS_FOR_PNG 6,0 0,3 0,42 15.5,41.984755
    """

    default_type = "Airbridge Multi Face"
    include_bumps = Param(pdt.TypeBoolean, "Include bumps", True)

    def build(self):
        half_length = self.bridge_length / 2
        half_box_length = half_length + self.pad_length
        half_width = self.bridge_width / 2

        # Add base metal addition to second face
        shape = pya.DBox(-half_width, -half_box_length, half_width, half_box_length)
        self.cell.shapes(self.get_layer("base_metal_addition", 1)).insert(shape)

        # Add ground grid avoidance to second face
        self.add_protection(shape.enlarged(self.margin, self.margin), 1)

        # Flip-chip bump
        if self.include_bumps:
            bump_dist = half_length + self.pad_length / 2
            bump = self.add_element(FlipChipConnectorDc)
            self.insert_cell(bump, pya.DTrans(pya.DPoint(0.0, bump_dist)))
            self.insert_cell(bump, pya.DTrans(pya.DPoint(0.0, -bump_dist)))

        # refpoints for connecting to waveguides
        self.add_port("a", pya.DPoint(0, half_length), pya.DVector(0, 1))
        self.add_port("b", pya.DPoint(0, -half_length), pya.DVector(0, -1))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import numpy

from kqcircuits.pya_resolver import pya

from kqcircuits.chips.chip import Chip
from kqcircuits.test_structures.stripes_test import StripesTest
from kqcircuits.test_structures.stripes_test_increasing_width import StripesTestIncreasingWidth
from kqcircuits.test_structures.cross_test import CrossTest
from kqcircuits.util.parameters import Param, pdt


class LithographyTest(Chip):
    """Optical lithography test chip.

    Consists of StripesTest cells with different parameters.
    """
    stripe_test = Param(pdt.TypeString, "stripe test type", "stripe_test",
                        choices=["stripe_test", "stripe_test_increasing_width"])

    def build(self):
        cell_horizontal_1, cell_vertical_1, cell_diagonal_1 = self.create_pattern(num_stripes=20, length=100,
                                                                                  min_width=1,
                                                                                  max_width=15, step=1, spacing=1,
                                                                                  face_id=self.face_ids[0])
        cell_horizontal_2, cell_vertical_2, cell_diagonal_2 = self.create_pattern(num_stripes=20, length=100,
                                                                                  min_width=1,
                                                                                  max_width=15, step=1, spacing=5,
                                                                                  face_id=self.face_ids[0])
        self.insert_cell(cell_horizontal_1, pya.DCplxTrans(1, 0, False, 2000, 6500))
        self.insert_cell(cell_horizontal_1, pya.DCplxTrans(1, 0, False, 3000, 6500))
        self.insert_cell(cell_horizontal_2, pya.DCplxTrans(1, 0, False, 4000, 6500))
        self.insert_cell(cell_horizontal_2, pya.DCplxTrans(1, 0, False, 6000, 6500))
        self.insert_cell(cell_vertical_1, pya.DCplxTrans(1, 0, False, 1500, 5700))
        self.insert_cell(cell_vertical_1, pya.DCplxTrans(1, 0, False, 5500, 5700))
        self.insert_cell(cell_vertical_2, pya.DCplxTrans(1, 0, False, 1500, 3800))
        self.insert_cell(cell_vertical_2, pya.DCplxTrans(1, 0, False, 5500, 3800))
        self.insert_cell(cell_diagonal_1, pya.DCplxTrans(1, 0, False, 400, 1500))
        self.insert_cell(cell_diagonal_1, pya.DCplxTrans(1, 0, False, 1500, 1500))
        self.insert_cell(cell_diagonal_2, pya.DCplxTrans(1, 0, False, 3100, 1500))
        self.insert_cell(cell_diagonal_2, pya.DCplxTrans(1, 0, False, 6100, 1500))

    def create_pattern(self, num_stripes, length, min_width, max_width, step, spacing, face_id):
        first_stripes_width = 2 * num_stripes * min_width
        cell_horizontal = self.layout.create_cell("Stripes")
        cell_vertical = self.layout.create_cell("Stripes")
        cell_diagonal = self.layout.create_cell("Stripes")

        for i, width in enumerate(numpy.arange(min_width, max_width + 0.1 * step, step)):
            n_stripes = num_stripes
            if self.stripe_test == "stripe_test":
                stripes_cell = self.add_element(StripesTest, num_stripes=n_stripes, stripe_width=width,
                                                stripe_length=length, stripe_spacing=spacing * width, face_ids=[face_id]
                                                )
                stripe_row_width = n_stripes * (width + spacing * width) - spacing * width
            elif self.stripe_test == "stripe_test_increasing_width":
                n_stripes = int(num_stripes * 0.4)
                stripes_cell = self.add_element(StripesTestIncreasingWidth, num_stripes=n_stripes,
                                                min_stripe_width=width, stripe_step=width,
                                                stripe_length=length, stripe_spacing=spacing * width, face_ids=[face_id]
                                                )
                stripe_row_width = sum([spacing * width + (i+1)*width for i in range(n_stripes)])
            # calculate the number of cross alignment marker
            if (stripe_row_width - 45) % 100 < 50:
                num_crosses = (stripe_row_width - 45) // 100 + 1
            else:
                num_crosses = (stripe_row_width - 45) // 100 + 2

            cross_cell = self.add_element(CrossTest, num_crosses=int(num_crosses), cross_spacing=100,
                                          face_ids=[face_id])

            # horizontal
            cell_horizontal.insert(pya.DCellInstArray(stripes_cell.cell_index(),
                                                      pya.DCplxTrans(1, 0, False, 0, 2 * i * length +
                                                                     first_stripes_width)))
            cell_horizontal.insert(pya.DCellInstArray(cross_cell.cell_index(),
                                                      pya.DCplxTrans(1, 0, False, 45, 2 * i * length +
                                                                     first_stripes_width -
                                                                     CrossTest.cross_length * 5 / 2)))
            # vertical
            cell_vertical.insert(pya.DCellInstArray(stripes_cell.cell_index(),
                                                    pya.DCplxTrans(1, 90, False, 2 * i * length + length +
                                                                   first_stripes_width, 0)))
            cell_vertical.insert(pya.DCellInstArray(cross_cell.cell_index(),
                                                    pya.DCplxTrans(1, 90, True, 2 * i * length + length +
                                                                   first_stripes_width - length -
                                                                   CrossTest.cross_length * 5 / 2, +45)))
            # diagonal
            diag_offset = 0  # 2*num_stripes*width/numpy.sqrt(8)
            cell_diagonal.insert(pya.DCellInstArray(stripes_cell.cell_index(),
                                                    pya.DCplxTrans(1, -45, False, 500 + i * length - diag_offset,
                                                                   500 + i * length + diag_offset)))

        return cell_horizontal, cell_vertical, cell_diagonal



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

"""
    Functions to replace SQUIDs in existing design files.

    Typical usage in a macro::

            top_cell = KLayoutView(current=True).active_cell
            replace_squids(top_cell, "MySQUID", "junction_width", 0.5, 0.1)  # a parameter sweep
            replace_squid(top_cell, "QB_2", "MySQUID", mirror=True)          # replace an individual SQUID
"""

from os import path
import logging
from kqcircuits.pya_resolver import pya
from kqcircuits.junctions import junction_type_choices
from kqcircuits.junctions.junction import Junction
from kqcircuits.chips.chip import Chip


def replace_squids(cell, junction_type, parameter_name, parameter_start, parameter_step, parameter_end=None):
    """Replaces squids by code generated squids with the given parameter sweep.

    All squids below top_cell in the cell hierarchy are removed. The number of code
    generated squids may be limited by the value of parameter_end.

    Args:
        cell (Cell): The cell where the squids to be replaced are
        junction_type: class name of the code generated squid that replaces the other squids
        parameter_name (str): Name of the parameter to be swept
        parameter_start: Start value of the parameter
        parameter_step: Parameter value increment step
        parameter_end: End value of the parameter. If None, there is no limit for the parameter value, so that all
            squids are replaced

    """
    layout = cell.layout()
    parameter_value = parameter_start
    junction_types = [choice if isinstance(choice, str) else choice[1] for choice in junction_type_choices]

    old_squids = []  # list of tuples (squid instance, squid dtrans with respect to cell, old name)

    def recursive_replace_squids(top_cell_inst, combined_dtrans):
        """Appends to old_squids all squids in top_cell_inst or any instance below it in hierarchy."""
        # cannot use just top_cell_inst.cell due to klayout bug, see
        # https://www.klayout.de/forum/discussion/1191/cell-shapes-cannot-call-non-const-method-on-a-const-reference
        top_cell = layout.cell(top_cell_inst.cell_index)
        for subcell_inst in top_cell.each_inst():
            subcell_name = subcell_inst.cell.name
            if subcell_name in junction_types:
                old_squids.append((subcell_inst, combined_dtrans*subcell_inst.dtrans, subcell_name))
            else:
                recursive_replace_squids(subcell_inst, combined_dtrans*subcell_inst.dtrans)

    for inst in cell.each_inst():
        if inst.cell.name in junction_types:
            old_squids.append((inst, inst.dtrans, inst.cell.name))
        recursive_replace_squids(inst, inst.dtrans)

    # sort left-to-right and bottom-to-top
    old_squids.sort(key=lambda squid: (squid[1].disp.x, squid[1].disp.y))

    for (inst, dtrans, name) in old_squids:
        if (parameter_end is None) or (parameter_value <= parameter_end):
            # create new squid at old squid's position
            parameters = {parameter_name: parameter_value}
            squid_cell = Junction.create(layout, junction_type=junction_type, face_ids=inst.pcell_parameter("face_ids"),
                                      **parameters)
            cell.insert(pya.DCellInstArray(squid_cell.cell_index(), dtrans))
            logging.info("Replaced squid \"%s\" with dtrans=%s by a squid \"%s\" with %s=%s.",
                         name, dtrans, junction_type, parameter_name, parameter_value)
            parameter_value += parameter_step
        # delete old squid
        inst.delete()

def replace_squid(top_cell, inst_name, junction_type, mirror=False, squid_index=0, **params):
    """Replaces a SQUID by the requested alternative in the named instance.

    Replaces the SQUID(s) in the sub-element(s) named ``inst_name`` with other SQUID(s) of
    ``junction_type``. The necessary SQUID parameters are specified in ``params``. If ``inst_name`` is
    a Test Structure then ``squid_index`` specifies which SQUID to change.

    Args:
        top_cell: The top cell with SQUIDs to be replaced
        inst_name: Instance name of PCell containing the SQUID to be replaced
        junction_type: Name of SQUID Class or .gds/.oas file
        mirror: Mirror the SQUID along its vertical axis
        squid_index: Index of the SQUID to be replaced within a Test Structure
        **params: Extra parameters for the new SQUID
    """

    def find_cells_with_squids(chip, inst_name):
        """Returns the container cells in `chip` called `inst_name`"""
        cells = []
        layout = chip.layout()
        for inst in chip.each_inst():
            if inst.property("id") == inst_name:
                cells.append((chip, inst))
            elif isinstance(inst.pcell_declaration(), Chip):  # recursively look for more chips
                cells += find_cells_with_squids(layout.cell(inst.cell_index), inst_name)
        return cells

    cells = find_cells_with_squids(top_cell, inst_name)
    if not cells:
        logging.warning(f"Could not find anything named '{inst_name}'!")

    layout = top_cell.layout()
    file_cell = None
    if junction_type.endswith(".oas") or junction_type.endswith(".gds"):  # try to load from file
        if not path.exists(junction_type):
            logging.warning(f"No file found at '{path.realpath(junction_type)}!")
            return
        load_opts = pya.LoadLayoutOptions()
        load_opts.cell_conflict_resolution = pya.LoadLayoutOptions.CellConflictResolution.RenameCell
        layout.read(junction_type, load_opts)
        file_cell = layout.top_cells()[-1]
        file_cell.name = f"Junction Library.{file_cell.name}"

    for (chip, inst) in cells:
        orig_trans = inst.dcplx_trans
        ccell = inst.layout().cell(inst.cell_index)

        if ccell.is_pcell_variant():  # make copy if used elsewhere
            dup = ccell.dup()
            dup.set_property("id", inst.property("id"))
            inst.delete()
            chip.insert(pya.DCellInstArray(dup.cell_index(), orig_trans), dup.prop_id)
            ccell = dup

        squids = [sq for sq in ccell.each_inst() if sq.cell.qname().find("Junction Library") != -1]
        squids.sort(key=lambda q: q.property("squid_index"))
        if not squids or squid_index >= len(squids) or squid_index < 0:
            logging.warning(f"No SQUID found in '{inst_name}' or squid_index={squid_index} is out of range!")
            continue
        old_squid = squids[squid_index]
        if old_squid.is_pcell():
            params = {"face_ids": old_squid.pcell_parameter("face_ids"), **params}
        trans = old_squid.dcplx_trans * pya.DCplxTrans.M90 if mirror else old_squid.dcplx_trans
        squid_pos = (orig_trans * trans).disp
        logging.info(f"Replaced SQUID of '{inst_name}' with {junction_type} at {squid_pos}.")
        old_squid.delete()
        if file_cell:
            new_squid = ccell.insert(pya.DCellInstArray(file_cell.cell_index(), trans))
        else:
            new_squid = Junction.create(layout, junction_type=junction_type, **params)
            new_squid = ccell.insert(pya.DCellInstArray(new_squid.cell_index(), trans))
        new_squid.set_property("squid_index", squid_index)

def convert_cells_to_static(layout):
    """Converts all cells in the layout to static. """

    converted_cells = {}

    # convert the cells to static
    for cell in layout.each_cell():
        if cell.is_library_cell():
            cell_idx = cell.cell_index()
            new_cell_idx = layout.convert_cell_to_static(cell_idx)
            if new_cell_idx != cell_idx:
                converted_cells[cell_idx] = new_cell_idx

    # translate the instances
    for cell in layout.each_cell():
        for inst in cell.each_inst():
            if inst.cell_index in converted_cells:
                inst.cell_index = converted_cells[inst.cell_index]

    # delete the PCells
    for cell_idx in converted_cells:
        layout.delete_cell(cell_idx)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya


def add_parameters_from(cls, /, *param_names, **param_with_default_value):
    """Decorator function to add parameters to the decorated class.

    Only the named parameters are added or changed. Use a starting wildcard (``"*"``) argument to
    get all parameters of ``cls``. For simplicity, if nothing is specified it gets all parameters.

    Parameters in ``param_names`` after the starting ``"*"`` will be *excluded*. The ``"*"`` is also
    useful together with ``param_with_default_value`` to get all parameters but change some.

    Args:
        cls: the class to take parameters from
        *param_names: parameter names to take (or remove if '*' is the first)
        **param_with_default_value: dictionary of parameter names and new default values
    """

    invert = (not param_names and not param_with_default_value)
    if param_names and param_names[0] == "*":
        param_names = param_names[1:]
        invert = True

    unknown = (set(param_names) | set(param_with_default_value.keys())) - set(cls.get_schema().keys())
    if unknown:
        raise ValueError(f"Parameter(s) {unknown} not available in '{cls.__name__}'")

    def _decorate(obj):
        for name, p in cls.get_schema().items():
            if name in param_with_default_value:
                # Redefine the Param object, if needed, because multiple elements may refer to it
                if param_with_default_value[name] != p.default:
                    p = Param(p.data_type, p.description, param_with_default_value[name], **p.kwargs)
            elif invert ^ (name not in param_names):
                continue
            setattr(obj, name, p)
            p.__set_name__(obj, name)
        return obj

    return _decorate


def add_parameter(cls, name, **change):
    """Decorator function to add a single parameter to the decorated class.

    Makes it possible to have fine-grained control over the Parameter's properties. Particularly,
    changing the "hidden" or the "choices" property of the parameter.

    Args:
        cls: the class to take parameters from
        name: name of the re-used parameter
        **change: dictionary of properties to change, like ``hidden=True``
    """

    schema = cls.get_schema()
    if name not in schema:
        raise ValueError(f"Parameter {name} not available in '{cls.__name__}'")

    def _decorate(obj):
        p = schema[name]
        if change:  # Redefine the parameter in case of any change
            kwargs = {**p.kwargs, "description":p.description, "default": p.default, **change}
            p = Param(p.data_type, **kwargs)
        setattr(obj, name, p)
        p.__set_name__(obj, name)
        return obj

    return _decorate


class pdt:  # pylint: disable=invalid-name
    """A namespace for pya.PCellParameterDeclaration types."""
    TypeDouble = pya.PCellParameterDeclaration.TypeDouble
    TypeInt = pya.PCellParameterDeclaration.TypeInt
    TypeList = pya.PCellParameterDeclaration.TypeList
    TypeString = pya.PCellParameterDeclaration.TypeString
    TypeNone = pya.PCellParameterDeclaration.TypeNone
    TypeShape = pya.PCellParameterDeclaration.TypeShape
    TypeBoolean = pya.PCellParameterDeclaration.TypeBoolean
    TypeLayer = pya.PCellParameterDeclaration.TypeLayer
    TypeNone = pya.PCellParameterDeclaration.TypeNone


class Param:
    """PCell parameters as Element class attributes.

    This should be used for defining PCell parameters in Element subclasses. The attributes of Param are same
    as ``pya.PCellParameterDeclaration``'s attributes, except for:

        * ``data_type``: same as ``type`` in ``pya.PCellParameterDeclaration``
        * ``choices``: List of (description, value) tuples or plain ``str`` values that are used as description too.
        * ``docstring``: Longer description of the parameter that gets used by Sphinx to generate API docs.
    """

    _index = {} # A private dictionary of parameter dictionaries indexed by owner classes' name

    @classmethod
    def get_all(cls, owner):
        """Get all parameters of given owner.

        Args:
            owner: get all parameters of this class

        Returns:
            a name-to-Param dictionary of all parameters of class `owner` or an empty one if it has none.
        """

        owner_name = f'{owner.__module__}.{owner.__qualname__}'
        if owner_name in cls._index:
            return cls._index[owner_name]
        else:
            return {}

    def __init__(self, data_type, description, default, **kwargs):
        self.data_type = data_type
        self.description = description
        self.default = default
        self.kwargs = kwargs

    def __set_name__(self, owner, name):
        self.name = name
        owner_name = f'{owner.__module__}.{owner.__qualname__}'
        if owner_name not in self._index:
            self._index[owner_name] = {}
        self._index[owner_name][name] = self

    def __get__(self, obj, objtype):
        if obj is None or not hasattr(obj, "_param_values") or obj._param_values is None:
            return self.default
        if hasattr(obj, "_param_value_map"):    # Element
            return obj._param_values[obj._param_value_map[self.name]]
        else:                                   # Simulation
            return obj._param_values[self.name]

    def __set__(self, obj, value):
        if not hasattr(obj, "_param_values") or obj._param_values is None:
            obj._param_values = {}
        if hasattr(obj, "_param_value_map"):
            obj._param_values[obj._param_value_map[self.name]] = value
        else:
            obj._param_values[self.name] = value



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import math
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.element import Element


class MaskMarkerFc(Element):
    """The PCell declaration for a MaskMarkerFc.

    Mask alignment marker for flip-chip masks
    """

    window = Param(pdt.TypeBoolean, "Window in airbridge flyover and UBM layer", False)
    arrow_number = Param(pdt.TypeInt, "Number of arrow pairs in the marker", 3)

    @staticmethod
    def create_cross(arm_length, arm_width):

        m = arm_length / 2
        n = arm_width / 2

        cross = pya.DPolygon([
            pya.DPoint(-n, -m),
            pya.DPoint(-n, -n),
            pya.DPoint(-m, -n),
            pya.DPoint(-m, n),
            pya.DPoint(-n, n),
            pya.DPoint(-n, m),
            pya.DPoint(n, m),
            pya.DPoint(n, n),
            pya.DPoint(m, n),
            pya.DPoint(m, -n),
            pya.DPoint(n, -n),
            pya.DPoint(n, -m)
        ])

        return cross

    def build(self):

        arrow = pya.DPolygon([
            pya.DPoint(0, 0),
            pya.DPoint(-20, 20),
            pya.DPoint(-11, 25),
            pya.DPoint(-5, 19),
            pya.DPoint(-5, 62),
            pya.DPoint(5, 62),
            pya.DPoint(5, 19),
            pya.DPoint(11, 25),
            pya.DPoint(20, 20)

        ])

        layer_gap = self.get_layer("base_metal_gap_wo_grid")
        layer_pads = self.get_layer("airbridge_pads")
        layer_flyover = self.get_layer("airbridge_flyover")
        layer_ubm = self.get_layer("underbump_metallization")
        layer_indium_bump = self.get_layer("indium_bump")
        layer_protection = self.get_layer("ground_grid_avoidance")

        def insert_to_main_layers(shape):
            self.cell.shapes(layer_gap).insert(shape)
            if not self.window:
                self.cell.shapes(layer_flyover).insert(shape)

        # protection for the box
        protection_box = pya.DBox(
            pya.DPoint(200, 350),
            pya.DPoint(-200, -350)
        )
        self.cell.shapes(layer_protection).insert(protection_box)
        negative_layer = pya.Region([protection_box.to_itype(self.layout.dbu)])

        # crosses

        arm_widths = [20, 35, 70]
        arm_lengths = [70, 124, 250]
        offset_width = [30, 43, 85]
        offset_length = [80, 132, 265]

        shift = pya.DPoint(0, 250)
        dislocation = 0
        for i, (cross_width, cross_length) in enumerate(zip(arm_widths, arm_lengths)):
            if i != 0:
                dislocation = arm_lengths[i - 1] + arm_lengths[i] + 100
            shift += pya.DPoint(0, -dislocation / 2)
            inner_shapes = pya.DCplxTrans(1, 0, False,
                                          pya.DVector(shift)) * self.create_cross(offset_length[i], offset_width[i])
            insert_to_main_layers(inner_shapes)
            inner_corner_shape = pya.DCplxTrans(1, 0, False,
                                                pya.DVector(shift)) * self.create_cross(cross_length, cross_width)
            for layer_insert in [layer_pads]:
                self.cell.shapes(layer_insert).insert(inner_corner_shape)
            negative_offset = pya.DCplxTrans(1, 0, False,
                                             pya.DVector(shift)) * self.create_cross(arm_lengths[i], arm_widths[i])
            negative_layer -= pya.Region([negative_offset.to_itype(self.layout.dbu)])

            inner_shapes_offset = pya.DCplxTrans(1, 0, False,
                                                 pya.DVector(shift)) * self.create_cross(arm_lengths[i], arm_widths[i])
            self.cell.shapes(layer_indium_bump).insert(inner_shapes_offset)
        self.cell.shapes(layer_ubm).insert(negative_layer)
        # marker arrow
        for i in range(self.arrow_number):
            for j in [-1, 1]:
                arrows_shape = pya.DCplxTrans(1, 0, j != 1,
                                              j * pya.DVector(0, 200 * i + 400)) * arrow
                insert_to_main_layers(arrows_shape)
                for layer_insert in [layer_pads, layer_indium_bump]:
                    self.cell.shapes(layer_insert).insert(arrows_shape)

    @classmethod
    def get_marker_locations(cls, cell_marker, **kwargs):
        # set markers to the edge clearance
        wafer_center_x = kwargs.get('wafer_center_x', 0)
        wafer_center_y = kwargs.get('wafer_center_y', 0)
        wafer_rad = kwargs.get('wafer_rad', 75000)
        edge_clearance = kwargs.get('edge_clearance', 1000)
        margin = kwargs.get('box_margin', 1000)
        _h = cell_marker.dbbox().height()
        _w = cell_marker.dbbox().width()
        coordinate = math.sqrt((wafer_rad - edge_clearance) ** 2 - (_h / 2 + margin) ** 2)
        return [
            pya.DTrans(wafer_center_x - coordinate + (margin + _w/2), wafer_center_y) * pya.DTrans.M90,
            pya.DTrans(wafer_center_x + coordinate - (margin + _w/2), wafer_center_y) * pya.DTrans.R0]

    @classmethod  # TODO: this is a direct copy from marker.py Will be fixed in future issue
    def get_marker_region(cls, inst, **kwargs):
        margin = kwargs.get('box_margin', 1000)
        return pya.Region(inst.bbox()).extents(margin / inst.cell.layout().dbu)



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from kqcircuits.simulations.simulation import Simulation
from kqcircuits.simulations.port import EdgePort
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.flip_chip_connectors.flip_chip_connector_dc import FlipChipConnectorDc


class WaveGuidesSim(Simulation):

    cpw_length = Param(pdt.TypeDouble, "Waveguide length", 100, unit="μm")
    n_guides = Param(pdt.TypeInt, "Number of guides", 5)
    spacing = Param(pdt.TypeDouble, "Parallel spacing", 100, unit="μm")
    guide_face_id = Param(pdt.TypeString, "Guide face id", '1t1')
    add_bumps = Param(pdt.TypeBoolean, "Add ground bumps", False)
    port_termination_end = Param(pdt.TypeBoolean, "Port termination end", True)
    use_edge_ports = Param(pdt.TypeBoolean, "Use edge ports", True)
    etch_opposite_face = Param(pdt.TypeBoolean, "Remove the whole opposite face metal if flip chip", False)


    def build(self):
        self.produce_guides()
        if self.add_bumps:
            self.produce_ground_bumps()

    def produce_guides(self):
        cpw_length = self.cpw_length
        n_guides = self.n_guides
        a = self.a
        b = self.b
        spacing = self.spacing
        tot_y = (n_guides-1)*spacing
        guide_face_id = self.guide_face_id
        face_id = {
            '1t1': 0,
            '2b1': 1
        }

        for i in range(n_guides):
            y_pos = i*spacing-tot_y/2.
            p1 = pya.DPoint(-cpw_length/2, y_pos)
            p2 = pya.DPoint(cpw_length/2., y_pos)
            p0 = pya.DPoint(0, y_pos)
            # waveguide_cell = self.add_element(WaveguideCoplanar, path=pya.DPath([p1, p2], 0),
            #                                   face_ids=[guide_face_id])
            if self.use_edge_ports:
                wg_cell = self.add_element(WaveguideCoplanar, path=pya.DPath([p1, p2], 0), term1=0,
                                           term2=0, face_ids=[guide_face_id])
                self.insert_cell(wg_cell)
                self.ports.append(EdgePort(i+1, p1, face=face_id[guide_face_id]))
                if self.port_termination_end:
                    self.ports.append(EdgePort(n_guides+i+1, p2, face=face_id[guide_face_id]))
            else:
                if self.port_termination_end:
                    self.produce_waveguide_to_port(p0, p1, i+1, waveguide_length=cpw_length/2.,
                                                   a=a, b=b, face=face_id[guide_face_id])
                    self.produce_waveguide_to_port(p0, p2, n_guides+i+1, waveguide_length=cpw_length/2.,
                                                   a=a, b=b, face=face_id[guide_face_id])
                else:
                    self.produce_waveguide_to_port(p0, p1, i+1, waveguide_length=cpw_length/2.,
                                                   a=a, b=b, face=face_id[guide_face_id])
                    wg_cell = self.add_element(WaveguideCoplanar, path=pya.DPath([p0, p2], 0), term1=0,
                                               term2=self.b, face_ids=[guide_face_id])
                    self.insert_cell(wg_cell)

            if self.etch_opposite_face:
                region = pya.Region(self.box.to_itype(self.layout.dbu))
                self.cell.shapes(self.get_layer("base_metal_gap_wo_grid", face_id=1)).insert(region)

    def produce_ground_bumps(self):
        n_guides = self.n_guides
        spacing = self.spacing
        tot_y = (n_guides-1)*spacing
        bump = self.add_element(FlipChipConnectorDc)

        for i in range(n_guides-1):
            y_pos = i*spacing-tot_y/2.+spacing/2.
            self.insert_cell(bump, pya.DCplxTrans(1, 0, False, pya.DVector(0, y_pos)))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
import logging
from os import cpu_count
from time import perf_counter

from kqcircuits.pya_resolver import pya


class AreaReceiver(pya.TileOutputReceiver):
    """ Class for handling and storing output from :class:`TilingProcessor` """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.area = 0.0

    def put(self, ix, iy, tile, obj, dbu, clip):
        """ Function called by :class:`TilingProcessor` on output """
        #pylint: disable=unused-argument
        logging.debug(f"Area for tile {ix},{iy}: {obj} ({dbu})")
        self.area = obj * (dbu * dbu)  # report as um^2


def get_area_and_density(cell: pya.Cell, layer_infos=None):
    """ Get total area and density :math:`\\rho=\\frac{area}{bbox.area}` of all layers.

    Args:
        cell: target cell to get area from
        layer_infos: list of ``LayerInfo`` to get area for, or None to get area for all layers.

    Returns:
        tuple: tuple containing lists of

          * layer names as str
          * total area as float
          * density between 0 and 1 as float

    """
    start_time = perf_counter()
    layout = cell.layout()

    tp = pya.TilingProcessor()
    tp.threads = cpu_count()
    tp.tile_size = (2000, 2000)  # microns
    if layer_infos is None:
        layer_infos = list(layout.layer_infos())
    layer_areas = [AreaReceiver() for _ in layer_infos]
    layer_bboxes = [AreaReceiver() for _ in layer_infos]

    for layer_info, area_receiver, bbox_receiver in zip(layer_infos, layer_areas, layer_bboxes):
        name = f"_{layer_info.name}"  # if `name` starts with a number, tp.execute() fails, so we add an underscore
        area, bbox = name + '_area', name + '_bbox'
        tp.input(name, cell.begin_shapes_rec(layout.layer(layer_info)))
        tp.output(area, area_receiver)
        tp.output(bbox, bbox_receiver)
        tp.queue(f"_output({area}, {name}.area)")
        tp.queue(f"_output({bbox}, {name}.bbox.area)")
    tp.execute("Calculate polygon and bounding box area")

    areas = [area.area for area in layer_areas]
    bboxes = [bbox.area for bbox in layer_bboxes]
    layer_names = [layer_info.name for layer_info in layer_infos]
    densities = [area / bbox if bbox != 0.0 else 0.0 for area, bbox in zip(areas, bboxes)]
    if len(areas) > 0:
        logging.info(f'For cell {cell.name} got layer areas: {areas}')
        logging.info(f'Area calculation took {perf_counter() - start_time:.1f} seconds')

    return layer_names, areas, densities



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

"""
    Helper module for building KLayout libraries.

    Typical usage example::

        from kqcircuits.elements.airbridges import Airbridge
        from kqcircuits.util.library_helper import load_libraries
        load_libraries(path=Airbridge.LIBRARY_PATH)
        cell = Airbridge.create(layout, **kwargs)
"""

import os
import re
import types
import inspect
import importlib
from pathlib import Path
import logging

from kqcircuits.defaults import SRC_PATHS, kqc_library_names, excluded_module_names
from kqcircuits.pya_resolver import pya


_kqc_libraries = {}  # dictionary {library name: (library, library path relative to kqcircuits)}

# Source directories not to be included in the library
_excluded_paths = (
    "__pycache__",
    "util",
    "simulations",
    "masks",
    "layer_config",
)

# modules NOT to be included in the library, (python file names without extension)
_excluded_module_names = (
    "__init__",
    "element",
    "qubit",
    "airbridge",
    "test_structure",
    "flip_chip_connector",
    "junction",
    "squid",
    "fluxline",
    "marker",
    "junction_test_pads",
    "tsv",
)

# Add user directories to SRC_PATHS when using Salt package.
if SRC_PATHS[0].parts[-4] == "salt":
    user_dirs = os.path.join(SRC_PATHS[0].parents[3], "python")
    for ud in os.listdir(user_dirs):
        SRC_PATHS.append(Path(os.path.join(user_dirs, ud)))


def load_libraries(flush=False, path=""):
    """Load all KQCircuits libraries from the given path.

    Args:
        flush: If True, old libraries will be deleted and new ones created. Otherwise old libraries will be used.
            (if old libraries exist)
        path: path (relative to SRC_PATH) from which the pcell classes and cells are loaded to libraries

    Returns:
         A dictionary of libraries that have been loaded, keys are library names and values are libraries.
    """

    if flush:
        delete_all_libraries()
        _kqc_libraries.clear()
        logging.debug("Deleted all libraries.")
    else:
        # if a library with the given path already exist, use it
        for _, lib_path in _kqc_libraries.values():
            if lib_path == path:
                return {key: value[0] for key, value in _kqc_libraries.items()}

    for cls in _get_all_pcell_classes(flush, path):

        library_name = cls.LIBRARY_NAME
        library_path = cls.LIBRARY_PATH

        library = pya.Library.library_by_name(library_name)  # returns only registered libraries
        if (library is None) or flush:
            if library_name in _kqc_libraries.keys():
                logging.debug("Using created library \"%s\".", library_name)
                library, _ = _kqc_libraries[library_name]
            else:
                # create a library, but do not register it yet
                logging.debug(f"Creating new library \"{library_name}\".")
                library = pya.Library()
                library.description = cls.LIBRARY_DESCRIPTION
                _kqc_libraries[library_name] = (library, library_path)
            _register_pcell(cls, library, library_name)

    # Libraries should be registered in dependency-order, otherwise reload will crash.
    for library_name in kqc_library_names:
        if library_name not in _kqc_libraries:
            continue
        library, _ = _kqc_libraries[library_name]
        _load_manual_designs(library_name)
        if library_name not in library.library_names():
            library.register(library_name)  # library must be registered only after all cells have been added to it

    return {key: value[0] for key, value in _kqc_libraries.items()}


def get_library_paths():
    """Returns a list of library paths under kqcircuits."""
    return (path for _, path in _kqc_libraries.values())


def delete_all_libraries():
    """Delete all KQCircuits libraries from KLayout memory."""
    for name in reversed(kqc_library_names):
        delete_library(name)


def delete_library(name=None):
    """Delete a KQCircuits library.

    Calls library.delete() and removes the library from _kqc_libraries dict. Also deletes any libraries which have this
    library as a dependency.

    Args:
        name: name of the library
    """
    if name is None:
        return
    library = pya.Library.library_by_name(name)
    if library is None:  # already deleted
        return

    # delete this library
    library.delete()
    if name in _kqc_libraries:
        _kqc_libraries.pop(name)
    if library._destroyed():
        logging.info(f"Successfully deleted library '{name}'.")
    else:
        raise SystemError("Failed to delete library '[]'.".format(name))


def element_by_class_name(class_name: str, library_path: str = "elements", library_name: str = "Element Library"):
    """
    Find Element class by class name from a library

    Args:
        class_name: Class name to look up
        library_path: Path to pass to load_libraries
        library_name: Name of the library

    Returns: Class of the element, or None if the element is not in the library
    """
    layout = load_libraries(path=library_path)[library_name].layout()
    for pcell_id in layout.pcell_ids():
        pcell_class = layout.pcell_declaration(pcell_id).__class__
        if pcell_class.__name__ == class_name:
            return pcell_class

    return None


def to_module_name(class_name=None):
    """Converts class name to module name.

    Converts PascalCase class name to module name
    with each word lowercase and separated by space.

    Args:
        class_name: Class name.

    Returns:
        A lowercase and spaced by word string.

        For example::
            > module_name = _to_module_name("QualityFactor")
            > print(module_name)
            "quality_factor"
    """
    try:
        _is_valid_class_name(class_name)
    except ValueError as e:
        logging.exception("Failed to convert class name to module name.")
        raise e
    words = re.sub(r"(?<!^)(?=[A-Z])", "_", class_name).split("_")
    return _join_module_words(words)


def to_library_name(class_name=None):
    """Converts class name to library name.

    Converts PascalCase class name to library name
    with each word titled and separated by space.
    Single letter words are attached to the following word.

    Args:
        class_name: Class name.

    Returns:
        A titled and spaced by word string which may be used as library name.

        For example::
            > library_name = to_library_name("QualityFactor")
            > print(library_name)
            "Quality Factor"
    """
    try:
        _is_valid_class_name(class_name)
    except ValueError as e:
        logging.exception("Failed to convert class name to library name.")
        raise e
    words = re.sub(r"(?<!^)(?=[A-Z])", "_", class_name).split("_")
    return _join_library_words(words)


# ********************************************************************************
# PRIVATE METHODS
# ********************************************************************************


def _register_pcell(pcell_class, library, library_name):
    """Registers the PCell to the library.

    Args:
        pcell_class: class of the PCell
        library: Library where the PCell is registered to
        library_name: name of the library
    """
    try:
        pcell_name = to_library_name(pcell_class.__name__)
        library.layout().register_pcell(pcell_name, pcell_class())
        logging.debug(f"Registered pcell [{pcell_name}] to library {library_name}.")
    except Exception:  # pylint: disable=broad-except
        logging.warning(
            f"Failed to register pcell in class {pcell_class} to library {library_name}.",
            exc_info=True
        )


def _load_manual_designs(library_name):
    """Loads .oas files to the library

    Args:
        library_name: name of the library
    """
    library, rel_path = _kqc_libraries[library_name]

    for src in SRC_PATHS:
        for path in src.rglob("{}/**/*.oas".format(rel_path)):
            library.layout().read(str(path.absolute()))


def _get_all_pcell_classes(reload=False, path="", skip_modules=False):
    """Returns all PCell classes in the given path.

    Args:
        reload: Boolean determining if the modules in kqcircuits should be reloaded.
        path: path (relative to SRC_PATH) from which the classes are searched
        skip_modules: Do not consider some pcells classes defined in ``excluded_module_names``.

    Returns:
        List of the PCell classes
    """
    pcell_classes = []

    for src in SRC_PATHS:
        pkg = src.parts[-1]

        if path == "":
            library_src_paths = [f for f in src.iterdir() if f.is_dir() and f.name not in _excluded_paths]
        else:
            library_src_paths = [src.joinpath(path)]

        skip_list = _excluded_module_names if not skip_modules else _excluded_module_names + excluded_module_names

        for library_src in library_src_paths:
            module_paths = library_src.rglob("*.py")
            for mp in module_paths:
                module_name = mp.stem
                if module_name in skip_list:
                    continue
                # Get the module path starting from the "pkg" directory below project root directory.
                import_path_parts = mp.parts[::-1][mp.parts[::-1].index(pkg)::-1]
                import_path = ".".join(import_path_parts)[:-3]  # the -3 is for removing ".py" from the path

                module = importlib.import_module(import_path)
                if reload:
                    importlib.reload(module)
                    logging.debug(f"Reloaded module '{module_name}'.")

                classes = _get_pcell_classes(module)
                if classes:
                    pcell_classes.append(classes[-1])

    return pcell_classes


def _get_pcell_classes(module=None):
    """Returns all PCell classes found in the module.

    Args:
        module: Module.

    Returns:
        Array of classes.
    """
    if module is None:
        return []
    class_list = []
    for member in inspect.getmembers(module, inspect.isclass):
        if member[1].__module__ == module.__name__:
            cls = _get_pcell_class(member[0], module)
            if cls is not None:
                class_list.append(cls)
    return class_list


def _get_pcell_class(name=None, module=None):
    """Returns class found for specified path and circuit type.

    Args:
        name: Path object for module.
        module: Module.

    Returns:
        The class if it is subclass of PCellDeclarationHelper, otherwise None.
    """
    if name is None or not isinstance(name, str):
        return None
    if module is None or not isinstance(module, types.ModuleType):
        return None
    value = getattr(module, name)
    if isinstance(value, type) and issubclass(value, pya.PCellDeclarationHelper):
        return value
    else:
        return None

def _is_valid_class_name(value=None):
    """Check if string value is valid PEP-8 compliant Python class name."""
    if value is None or not isinstance(value, str) or len(value) == 0:
        raise ValueError("Cannot convert nil or non-string class name '{}' to library name.".format(value))
    if re.fullmatch(r"[a-zA-Z_][a-zA-Z0-9_]*", value) is None:
        raise ValueError("Cannot convert invalid Python class name '{}' to library name.".format(value))
    if re.fullmatch(r"([A-Z][a-z0-9]*)+", value) is None:
        raise ValueError("PEP8 compliant class name '{}' must be PascalCase without underscores.".format(value))


def _join_module_words(words=None):
    """Join words to build module name.

    Joins words such that each word is lowercase and separated by underscore.
    Single letter words are attached to the following word.

    Args:
        words: List of word strings

    Returns:
        A string which may be used as module name.
    """
    words = _clean_words(words)
    words = [w.lower() for w in words]
    n = len(words)
    if n == 0:
        return ""
    name = words[0]
    for i in range(1, n):
        previous = words[i - 1]
        current = words[i]
        if len(previous) == 1:
            name += current
        else:
            name += "_" + current
    return name


def _join_library_words(words=None):
    """Join words to build library name.

    Joins words such that each word is titled and separated by space.
    Single letter words are attached to the following word.

    Args:
        words: List of word strings

    Returns:
        A string which may be used as library name.
    """
    words = _clean_words(words)
    n = len(words)
    if n == 0:
        return ""
    name = words[0].title()
    for i in range(1, n):
        previous = words[i - 1].title()
        current = words[i].title()
        if len(previous) == 1:
            name += current
        else:
            name += " " + current
    return name


def _clean_words(words=None):
    """Clean word list by removing None values, empty strings, and non-string values.

    Returns:
        Clean word list.
    """
    if words and isinstance(words, list):
        words = list(filter(None, words))
        words = list(filter(lambda item: isinstance(item, str), words))
        return list(filter(len, words))
    else:
        return []



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.elements.airbridges.airbridge import Airbridge
from kqcircuits.elements.element import Element
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.waveguide_coplanar_taper import WaveguideCoplanarTaper
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from


@add_parameters_from(Airbridge)
@add_parameters_from(WaveguideCoplanarTaper, "*", m2=5)
class AirbridgeConnection(Element):
    """The PCell declaration of an Airbridge with tapered waveguides in both ends.

     .. MARKERS_FOR_PNG 0,20 23,26 14,-15
    """

    bridge_gap_width = Param(pdt.TypeDouble, "Width of waveguide gap around the Airbridge", 12, unit="μm")
    with_side_airbridges = Param(pdt.TypeBoolean, "With airbridges on the sides", True)
    with_right_waveguide = Param(pdt.TypeBoolean, "With waveguide on right side", True)
    gap_between_bridges = Param(pdt.TypeDouble, "Inner distance between adjacent bridges", 20)
    waveguide_extra = Param(pdt.TypeDouble, "Waveguide extra length below airbridge", 0)

    def build(self):
        # Add airbridges
        ab = self.add_element(Airbridge)
        ab_inst, ab_ref = self.insert_cell(ab, pya.DTrans.R90)
        if self.with_side_airbridges:
            self.insert_cell(ab, pya.DTrans(1, False, 0,  self.bridge_width + self.gap_between_bridges))
            self.insert_cell(ab, pya.DTrans(1, False, 0, -self.bridge_width - self.gap_between_bridges))

        # Add left waveguide
        pad_a = self.bridge_width
        pad_b = self.bridge_gap_width
        wg_l_pos = ab_ref["port_a"] + pya.DVector(self.waveguide_extra, 0)
        taper_l_pos = ab_ref["port_a"] - pya.DVector(self.pad_length, 0)
        terminator_l = self.add_element(WaveguideCoplanar,
                                        path=[taper_l_pos, wg_l_pos],
                                        a=pad_a, b=pad_b,
                                        term1=0, term2=pad_b)
        self.insert_cell(terminator_l, pya.DTrans(0, 0))

        taper_l = self.add_element(WaveguideCoplanarTaper, a2=pad_a, b2=pad_b)
        taper_l_ref = self.get_refpoints(taper_l)
        taper_l_inst, _ = self.insert_cell(taper_l, pya.DTrans(taper_l_pos-taper_l_ref["port_b"].to_v()))
        self.copy_port("a", taper_l_inst)

        # Optionally, add right waveguide
        if self.with_right_waveguide:
            wg_r_pos = ab_ref["port_b"] - pya.DVector(self.waveguide_extra, 0)
            taper_r_pos = ab_ref["port_b"] + pya.DVector(self.pad_length, 0)
            terminator_r = self.add_element(WaveguideCoplanar,
                                            path=[wg_r_pos, taper_r_pos],
                                            a=pad_a, b=pad_b,
                                            term1=pad_b, term2=0)
            self.insert_cell(terminator_r, pya.DTrans(0, 0))

            taper_r = self.add_element(WaveguideCoplanarTaper, a=pad_a, b=pad_b)
            taper_r_ref = self.get_refpoints(taper_r)
            taper_r_inst, _ = self.insert_cell(taper_r, pya.DTrans(taper_r_pos-taper_r_ref["port_a"].to_v()))
            self.copy_port("b", taper_r_inst)
        else:
            wg_r_pos = ab_ref["port_b"]
            self.copy_port("b", ab_inst)

        # Add path
        path_airbridge = pya.DPath([wg_l_pos, wg_r_pos], self.bridge_width)
        self.cell.shapes(self.get_layer("waveguide_length")).insert(path_airbridge)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import json

from kqcircuits.pya_resolver import pya


class GeometryJsonEncoder(json.JSONEncoder):
    """JSON encoder that converts several pya D* types into JSON objects."""
    def default(self, o):
        if isinstance(o, pya.DPoint):
            return {"_pya_type": "DPoint", "x": o.x, "y": o.y}
        if isinstance(o, pya.DVector):
            return {"_pya_type": "DVector", "x": o.x, "y": o.y}
        if isinstance(o, pya.DBox):
            return {
                "_pya_type": "DBox",
                "p1": self.default(o.p1),
                "p2": self.default(o.p2)
            }
        if isinstance(o, pya.LayerInfo):
            return {"_pya_type": "LayerInfo", "layer": o.layer, "datatype": o.datatype}
        if isinstance(o, pya.DPath):
            return {
                "_pya_type": "DPath",
                "points": [self.default(p) for p in o.each_point()],
                "width": o.width
            }
        if isinstance(o, pya.DEdge):
            return {
                "_pya_type": "DEdge",
                "p1": self.default(o.p1),
                "p2": self.default(o.p2)
            }
        if isinstance(o, pya.DPolygon):
            return {
                "_pya_type": "DPolygon",
                "hull": [self.default(p) for p in o.each_point_hull()],
                "holes": [[self.default(p) for p in o.each_point_hole(i)] for i in range(o.holes())]
            }
        # Defer to standard json encoder for other cases
        return json.JSONEncoder.default(self, o)

class GeometryJsonDecoder(json.JSONDecoder):
    """JSON decoder that converts JSON objects into pya D* type objects."""
    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=GeometryJsonDecoder._decode_geometry, *args, **kwargs)

    @staticmethod
    def _decode_geometry(js):
        if "_pya_type" not in js:
            return js
        if js["_pya_type"] == "DPoint":
            return pya.DPoint(js["x"], js["y"])
        if js["_pya_type"] == "DVector":
            return pya.DVector(js["x"], js["y"])
        if js["_pya_type"] == "DBox":
            return pya.DBox(js["p1"], js["p2"])
        if js["_pya_type"] == "LayerInfo":
            return pya.LayerInfo(js["layer"], js["datatype"])
        if js["_pya_type"] == "DPath":
            return pya.DPath(js["points"], js["width"])
        if js["_pya_type"] == "DEdge":
            return pya.DEdge(js["p1"], js["p2"])
        if js["_pya_type"] == "DPolygon":
            polygon = pya.DPolygon(js["hull"])
            for hole in js["holes"]:
                polygon.insert_hole(hole)
            return polygon
        raise json.JSONDecodeError(f"_pya_type '{js['_pya_type']}' not currently deserializable: {js}", js, 0)

def encode_python_obj_as_dict(o):
    """Encodes a Python object into a JSON parseable dict

    Args:
        o - Python object
    """
    return json.loads(GeometryJsonEncoder().encode(o))

def decode_dict_as_python_obj(o):
    """Decodes a JSON parseable dict into a Python object

    Args:
        o - JSON parseable dict
    """
    return GeometryJsonDecoder().decode(json.dumps(o))
